#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from tvtk.api import tvtk

import numpy as np

# ------------------------------------------------------------------------------

from geolab.plotter.plotmanager import PlotManager

from geolab.plotter.pointsource import Points

from geolab.plotter.vectorsource import Vectors

from geolab.plotter.polylinesource import Polyline

# ------------------------------------------------------------------------------

'''-'''

__author__ = 'Davide Pellis'


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
#                               Vertices PlotManager
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


class BsplineCurvePlotManager(PlotManager):

    def __init__(self, scene_model=None):
        PlotManager.__init__(self, scene_model)

        self.scale = 1

        self.glyph_scale = 1

        self.vector_scale = 1

        self._vertex_data = None

        self.selected_vertices = []

        self.vertex_color = 'cornflower'

        self.control_color = 'cornflower'

        self.curve_color = 'cornflower'

        self.selection_color = 'yellow'

        self.view_mode = 'solid'

        self._r = None

        self._g = None

        self._bspline = None

        self._updating = True

        self._record = False

        self.__dimensions = (0)

        self.__counter = 0

    # -------------------------------------------------------------------------
    #                              Properties
    # -------------------------------------------------------------------------

    @property
    def type(self):
        return 'Points_plot_manager'

    @property
    def bspline(self):
        return self._bspline

    @bspline.setter
    def bspline(self, bspline):
        if bspline.type != 'Bspline_Curve':
            raise ValueError('*bspline* attribute must be a geometry.BsplineCurve!')
        self._bspline = bspline
        self._set_r()
        self._set_g()

    @property
    def geometry(self):
        return self._bspline

    @geometry.setter
    def geometry(self, bspline):
        self.bspline = bspline

    @property
    def r(self):
        return self._r * self.scale

    @property
    def g(self):
        return self._r * self._g * self.glyph_scale

    @property
    def v(self):
        return 12 * self._r * self.vector_scale

    @property
    def updating(self):
        return self._updating

    @updating.setter
    def updating(self, updating):
        if type(updating) == bool:
            self._updating = updating
        else:
            raise ValueError('the updating attribute must be a boolean!')

    @property
    def record(self):
        return self._record

    @record.setter
    def record(self, bool):
        self._record = bool

    @property
    def vertex_data(self):
        if self._vertex_data is None:
            self._vertex_data = np.zeros(self.mesh.V)
        return self._vertex_data

    @vertex_data.setter
    def vertex_data(self, vertex_data):
        self._vertex_data = vertex_data

    @property
    def object_selection_actors(self):
        return self.get_actor('virtual-points').actors

    # -------------------------------------------------------------------------
    #                              Size set
    # -------------------------------------------------------------------------

    def check_updates(self):
        updated = False
        if self.__dimensions != self.geometry.V:
            updated = True
        if updated:
            self._set_r()
            self._set_g()
            self._vertex_data = None
        self.__dimensions = (self.geometry.V)
        return updated

    def _set_r(self):
        if self.geometry is None or self.geometry.V == 0:
            self._r = None
        else:
            p_max = np.max(self.geometry.control_points, axis=0)
            p_min = np.min(self.geometry.control_points, axis=0)
            r = np.linalg.norm(p_max - p_min) / 70
            r = r / self.geometry.V ** 0.1
            if r == 0:
                r = np.linalg.norm(p_max) / 10 + 0.1
            self._r = r

    def _set_g(self):
        if self.geometry is None or self.geometry.V == 0:
            self._r = None
        else:
            p_max = np.max(self.geometry.control_points, axis=1)
            p_min = np.min(self.geometry.control_points, axis=1)
            g = np.linalg.norm(p_max - p_min) / 100
            if g == 0:
                g = np.linalg.norm(p_max) / 10 + 0.1
            self._g = g

    # -------------------------------------------------------------------------
    #                                Clear
    # -------------------------------------------------------------------------

    def clear(self, delete=True):
        PlotManager.clear(self)
        self.picker_off()
        self.selected_vertices = []
        self.remove_widgets()

    # -------------------------------------------------------------------------
    #                            Plot functions
    # -------------------------------------------------------------------------

    def initialize_plot(self):
        pass
        # self.plot_edges()

    def update_plot(self, **kwargs):
        self.check_updates()
        PlotManager.update_plot(self, **kwargs)
        self.plot_selected_vertices()

    def plot_vertices(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'vertices'
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        elif 'vertex_data' in kwargs:
            if self.record:
                kwargs['vertex_data'] = self.vertex_data
            else:
                self.vertex_data = kwargs['vertex_data']
        if kwargs['name'] not in self.sources or not self.updating:
            P = Points(self.geometry.vertices, **kwargs)
            self.add(P)
        else:
            self.update(**kwargs)

    def plot_control_polygon(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'control-polygon'
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r
        if 'color' not in kwargs:
            kwargs['color'] = self.control_color
        elif 'vertex_data' in kwargs:
            if self.record:
                kwargs['vertex_data'] = self.vertex_data
            else:
                self.vertex_data = kwargs['vertex_data']
        if kwargs['name'] not in self.sources or not self.updating:
            P = Polyline(self.geometry.control_polyline(), **kwargs)
            self.add(P)
        else:
            kwargs['polyline'] = self.geometry.control_polyline()
            self.update(**kwargs)

    def plot_curve(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'curve'
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        if kwargs['name'] not in self.sources or not self.updating:
            P = Polyline(self.geometry.curve_polyline(), **kwargs)
            self.add(P)
        else:
            kwargs['polyline'] = self.geometry.curve_polyline()
            self.update(**kwargs)

    def plot_glyph(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'glyph'
        if 'Vertices' not in kwargs:
            kwargs['vertices'] = self.mesh
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        if kwargs['name'] not in self.sources or not self.updating:
            P = Points(**kwargs)
            self.add(P)
        else:
            self.update(**kwargs)

    def plot_vectors(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'point-vectors'
        if 'anchor' not in kwargs:
            kwargs['anchor'] = self.geometry
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        if kwargs['name'] not in self.sources or not self.updating:
            P = Vectors(**kwargs)
            self.add(P)
            if 'scale_factor' not in kwargs:
                kwargs['scale_factor'] = self.v
                self.update(**kwargs)
        else:
            if 'scale_factor' not in kwargs:
                kwargs['scale_factor'] = self.v
            self.update(**kwargs)

    def hide_vertices(self):
        self.hide('vertices')

    def remove_vertices(self):
        self.remove('vertices')

    def hide_curve(self):
        self.hide('curve')

    def remove_curve(self):
        self.remove('curve')

    def hide_control_polygon(self):
        self.hide('control-polygon')

    def remove_control_polygon(self):
        self.remove('control-polygon')

    # -------------------------------------------------------------------------
    #                             Virtual plot
    # -------------------------------------------------------------------------

    def selection_on(self):
        self.virtual_vertices_on()
        return self.geometry.V * 8

    def selection_off(self):
        self.virtual_vertices_off()

    def virtual_vertices_on(self):
        if 'virtual-points' not in self.sources or not self.updating:
            M = Points(self.geometry.control_points,
                       name='virtual-points',
                       glyph_type='cube',
                       radius=self.r * 1.1,
                       opacity=0.001)
            self.add(M, pickable=True)
        else:
            self.update(name='virtual-points',
                        radius=self.r * 1.1)

    def virtual_vertices_off(self):
        self.hide('virtual-points')

    # --------------------------------------------------------------------------
    #                            Selection plot
    # --------------------------------------------------------------------------

    def highlight(self):
        radius = self.r * 1.1
        P = Points(self.geometry,
                   radius=radius,
                   color=self.selection_color,
                   shading=False,
                   name='highlight')
        self.add(P)

    def highlight_off(self):
        self.remove('highlight')

    def plot_selected_vertices(self):
        if len(self.selected_vertices) > 0:
            if 'selected-vertices' not in self.sources or not self.updating:
                if self.view_mode == 'wireframe' or self.view_mode == '3d':
                    glyph = 'wirefrane'
                    radius = None
                elif self.view_mode == 'solid':
                    radius = self.r * 1.1
                    glyph = 'sphere'
                P = Points(self.geometry.control_points[self.selected_vertices],
                           color=self.selection_color,
                           radius=radius,
                           shading=False,
                           glyph_type=glyph,
                           name='selected-vertices',
                           resolution=18)
                self.add(P)
            else:
                self.update(name='selected-vertices',
                            points=self.geometry.control_points[self.selected_vertices],
                            radius=self.r * 1.1)
        else:
            self.hide('selected-vertices')

    def hide_selected_vertices(self):
        self.hide('selected-vertices')

    # -------------------------------------------------------------------------
    #                               Selection
    # -------------------------------------------------------------------------

    def select_vertices(self):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = []

        def callback(p_id):

            if p_id == -1:
                return
            v = p_id // 6
            if v not in self.selected_vertices:
                self.selected_vertices.append(v)
            else:
                self.selected_vertices.remove(v)
            self.plot_selected_vertices()

        self.picker_callback(callback, mode='cell', name='virtual-points')

    def on_vertex_selection(self, vertex_callback):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = []

        def callback(p_id):
            if p_id == -1:
                return
            v = p_id // 6
            self.selected_vertices = [v]
            if vertex_callback is not None:
                vertex_callback(v)
            self.virtual_vertices_on()

        self.picker_callback(callback, mode='cell', name='virtual-points')

    def select_vertices_off(self):
        self.virtual_vertices_off()
        self.hide_selected_vertices()
        self.picker_off()
        self.selected_vertices = []

    def clear_selection(self):
        self.hide_selected_vertices()
        self.selected_vertices = []
        self.virtual_vertices_off()
        self.remove_widgets()

    def select_off(self):
        self.select_vertices_off()
        self.move_vertex_off()
        self.move_vertices_off()

    # --------------------------------------------------------------------------
    #                               Moving
    # --------------------------------------------------------------------------

    def move_vertex(self, vertex_index, interaction_callback=None,
                    end_callback=None):
        self.remove_widgets()
        S = tvtk.SphereWidget()
        point = self.geometry.control_points[vertex_index]
        S.center = point
        S.radius = self.r * 1.05
        S.representation = 'surface'
        P = Points(points=np.array([point]),
                   radius=self.r * 1.1,
                   color=self.selection_color,
                   shading=False,
                   name='widget-point')
        self.add(P)
        self.virtual_vertices_on()

        def i_callback(obj, event):
            # self.selected_vertices = []
            self.virtual_vertices_off()
            c = obj.GetCenter()
            center = np.array([c[0], c[1], c[2]])
            self.geometry.control_points[vertex_index, :] = center
            self.update(name='widget-point', points=center)
            if interaction_callback is not None:
                interaction_callback()

        S.add_observer("InteractionEvent", i_callback)

        def e_callback(obj, event):
            if end_callback is not None:
                end_callback()
            self.virtual_vertices_on()
            point = self.geometry.control_points[vertex_index]
            S.center = point
            self.update(name='widget-point', points=point)

        S.add_observer("EndInteractionEvent", e_callback)
        self.add_widget(S, name='vertex_handler')

    def move_vertex_off(self):
        self.remove_widgets()
        self.remove('widget-point')
        self.virtual_vertices_off()

    def move_vertices(self, interaction_callback=None, start_callback=None,
                      end_callback=None):
        self.select_off()
        self.virtual_vertices_on()

        def v_callback(p_id):
            self.hide_selected_vertices()
            if p_id == -1:
                return
            v = p_id // 6
            self.selected_vertices = [v]
            if start_callback is not None:
                try:
                    start_callback()
                except:
                    pass
            self.move_vertex(v, interaction_callback, end_callback)

        self.picker_callback(v_callback, mode='cell', name='virtual-points')

    def move_vertices_off(self):
        self.selected_vertices = []
        self.move_vertex_off()
        self.picker_off()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from traits.api import HasTraits, Instance, on_trait_change, Int

from traitsui.api import View, Item

from tvtk.pyface.scene_editor import SceneEditor

from mayavi.tools.mlab_scene_model import MlabSceneModel

from pyface.image_resource import ImageResource

# -----------------------------------------------------------------------------

from wsurf.geolab import utilities

from wsurf.geolab.plotter.geolabscene import GeolabScene

from wsurf.geolab.plotter.plotmanager import PlotManager

# -----------------------------------------------------------------------------

'''viewer.py: The viewer for plotter souce classes'''

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                    Viewer
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class Viewer(HasTraits):

    # -------------------------------------------------------------------------
    #                                  Traits
    # -------------------------------------------------------------------------

    scene = Instance(MlabSceneModel, ())

    editor = SceneEditor(scene_class=GeolabScene)

    # -------------------------------------------------------------------------
    #                                 Initialize
    # -------------------------------------------------------------------------

    def __init__(self, objects, **kwargs):
        HasTraits.__init__(self)

        self._plotmanager = PlotManager(scene_model=self.scene)

        self._generate_data(objects)

        self._size = kwargs.get('size', (800, 800))

        self._position = kwargs.get('position', None)

        self._background = kwargs.get('background', None)

        self._resizable = kwargs.get('resizable', True)
        
        self.magnification = kwargs.get('magnification', 4)

    @property
    def position(self):
        return self.scene._position

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, background):
        self._background = background

    # -------------------------------------------------------------------------
    #                                 Methods
    # -------------------------------------------------------------------------

    def start(self):

        view = View(Item('scene',
                         editor=SceneEditor(scene_class=GeolabScene),
                         show_label=False,
                         resizable=self._resizable,
                         height=self._size[1],
                         width=self._size[0],
                         ),
                    resizable=self._resizable,
                    title='GeoLab viewer',
                    icon=ImageResource('../icons/geolab_logo.png')
                    )

        self.configure_traits(view=view)

    def _generate_data(self, objects):
        self._plotmanager.add(objects, rename=True)
        
    @on_trait_change('scene.activated')
    def _scene_settings(self):
        if self._position is not None:
            self._plotmanager.set_view(self._position)
        if self.background is not None:
            self._plotmanager.background = self._background
        self.scene.magnification = self.magnification

# -----------------------------------------------------------------------------
#                                 View Function
# -----------------------------------------------------------------------------

def view(objects, **kwargs):
    viewer = Viewer(objects, **kwargs)
    viewer.start()
    
# -----------------------------------------------------------------------------
#                                 Save Function
# -----------------------------------------------------------------------------





from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

#------------------------------------------------------------------------------

from traits.api import HasTraits

#------------------------------------------------------------------------------

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                  Component
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class GeolabComponent(HasTraits):

    name = 'component'

    def __init__(self):
        HasTraits.__init__(self)
        self.__geolab = None

    @property
    def geolab(self):
        return self.__geolab

    @geolab.setter
    def geolab(self, geolab):
        self.__geolab = geolab
        self.__geolab.add_callback('object_change', self.geolab_object_change)
        self.__geolab.add_callback('object_changed', self.geolab_object_changed)
        self.__geolab.add_callback('object_save', self.geolab_object_save)
        self.__geolab.add_callback('object_saved', self.geolab_object_saved)
        self.__geolab.add_callback('object_added', self.geolab_object_added)
        self.__geolab.add_callback('object_add', self.geolab_object_add)

        #self.__geolab.handler.add_state_callback(self.geolab_set_state)

    #@property
    #def handler(self):
        #return self.geolab.handler

    def geolab_settings(self):
        pass

    def initialize_plot(self):
        pass

    def geolab_set_state(self, name):
        pass

    def geolab_object_change(self):
        pass

    def geolab_object_changed(self):
        pass

    def geolab_object_save(self, file_name):
        pass

    def geolab_object_saved(self, file_name):
        pass

    def geolab_object_added(self):
        pass

    def geolab_object_add(self):
        pass
    
    def geolab_closing(self):
        pass








from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, _sanity_check, _empty_output_check, _type_check, _create_mkl_sparse,
                                           _destroy_mkl_handle, matrix_descr, RETURN_CODES, _is_dense_vector,
                                           _out_matrix, _check_return_value, _is_allowed_sparse_format)

import numpy as np
import ctypes as _ctypes


def _sparse_dense_vector_mult(matrix_a, vector_b, scalar=1., transpose=False, out=None, out_scalar=None, out_t=None):
    """
    Multiply together a sparse matrix and a dense vector

    :param matrix_a: Left (A) matrix
    :type matrix_a: sp.spmatrix.csr, sp.spmatrix.csc
    :param vector_b: Right (B) vector with shape (N, ) or (N, 1)
    :type vector_b: np.ndarray
    :param scalar: A value to multiply the result matrix by. Defaults to 1.
    :type scalar: float
    :param transpose: Return AT (dot) B instead of A (dot) B.
    :type transpose: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: A (dot) B as a dense array
    :rtype: np.ndarray
    """

    output_shape = matrix_a.shape[1] if transpose else matrix_a.shape[0]
    output_shape = (output_shape,) if vector_b.ndim == 1 else (output_shape, 1)

    if _empty_output_check(matrix_a, vector_b):
        final_dtype = np.float64 if matrix_a.dtype != vector_b.dtype or matrix_a.dtype != np.float32 else np.float32
        return _out_matrix(output_shape, final_dtype, out_arr=out)

    mkl_a, dbl = _create_mkl_sparse(matrix_a)
    vector_b = vector_b.ravel()

    # Set functions and types for float or doubles
    output_ctype = _ctypes.c_double if dbl else _ctypes.c_float
    output_dtype = np.float64 if dbl else np.float32
    func = MKL._mkl_sparse_d_mv if dbl else MKL._mkl_sparse_s_mv

    output_arr = _out_matrix(output_shape, output_dtype, out_arr=out, out_t=out_t)

    ret_val = func(11 if transpose else 10,
                   scalar,
                   mkl_a,
                   matrix_descr(),
                   vector_b,
                   float(out_scalar) if out_scalar is not None else 1.,
                   output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)))

    # Check return
    _check_return_value(ret_val, func.__name__)

    _destroy_mkl_handle(mkl_a)

    return output_arr


def _sparse_dot_vector(mv_a, mv_b, cast=False, scalar=1., out=None, out_scalar=None):
    """
    Multiply a sparse matrix by a dense vector.
    The matrix must be CSR or CSC format.
    The vector must be (N,) or (N, 1) shape.
    Returns a dense vector of (N,) or (N, 1) shape (depending on vector)

    :param mv_a: Left (A) matrix or vector
    :type mv_a: np.ndarray, sp.spmatrix.csr, sp.spmatrix.csc
    :param mv_b: Right (B) matrix or vector
    :type mv_b: np.ndarray, sp.spmatrix.csr, sp.spmatrix.csc
    :param scalar: A value to multiply the result matrix by. Defaults to 1.
    :type scalar: float
    :param cast: Convert values to compatible floats if True. Raise an error if they are not compatible if False.
    Defaults to False.
    :type cast: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: A (dot) B as a dense matrix
    :rtype: np.ndarray
    """

    _sanity_check(mv_a, mv_b, allow_vector=True)
    mv_a, mv_b = _type_check(mv_a, mv_b, cast=cast)

    if not _is_allowed_sparse_format(mv_a) or not _is_allowed_sparse_format(mv_b):
        raise ValueError("Only CSR, CSC, and BSR-type sparse matrices are supported")
    elif _is_dense_vector(mv_b):
        return _sparse_dense_vector_mult(mv_a, mv_b, scalar=scalar, out=out, out_scalar=out_scalar)
    elif _is_dense_vector(mv_a) and out is None:
        return _sparse_dense_vector_mult(mv_b, mv_a.T, scalar=scalar, transpose=True).T
    elif _is_dense_vector(mv_a) and out is not None:
        _ = _sparse_dense_vector_mult(mv_b, mv_a.T, scalar=scalar, transpose=True,
                                      out=out.T, out_scalar=out_scalar, out_t=True)
        return out
    else:
        raise ValueError("Neither mv_a or mv_b is a dense vector")



#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from timeit import default_timer as timer

import numpy as np

from scipy import sparse

# ------------------------------------------------------------------------------

from wsurf.geolab.optimization import spsolve

from wsurf.geolab.optimization import spdot, hdot

# ------------------------------------------------------------------------------

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

class GuidedProjectionBase(object):

    epsilon = 0.001

    step = 1

    iterations = 10

    threshold = None

    make_residual = False

    plot_matrix = False

    step_control = False

    fairness_reduction = 0

    verbose = True

    _plot_matrix = False

    _weights = {}

    _N = 0

    _X = None

    _X0 = None

    _H = []

    _r = None

    _H0 = []

    _r0 = None

    _K = None

    _s = None

    _K0 = None

    _s0 = None

    _errors = {}

    _values = {}

    _messages = []

    _report = None

    _step = 1

    _counter = 0

    _initialize = True

    _reinitialize = True

    _constant_constraints = []

    _iterative_constraints = []

    _constraints = []

    _residual = None

    _residual_norm = None

    __timer = 0

    __t0 = 0

    __counter = 0

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------

    @property
    def N(self):
        return self._N

    @N.setter
    def N(self, N):
        if N != self._N:
            self._reinitialize = True
        self._N = N

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, X):
        self._X = X
        self._X0 = np.copy(X)

    @property
    def max_weight(self):
        return max(list(self._weights.values()))

    @property
    def weights(self):
        return self._weights

    @property
    def initialize(self):
        return self._initialize

    @initialize.setter
    def initialize(self, bool):
        self._initialize = bool

    @property
    def iteration(self):
        return self._counter

    # -------------------------------------------------------------------------
    #                                Workflows
    # -------------------------------------------------------------------------

    def reinitialize(self):
        self._reinitialize = True

    def _initialization(self):
        self._reset_counter()
        self.on_initialize()

    def initialization(self):
        self._initialization()
        self.reinitialization()

    def _reinitialization(self):
        self.on_reinitialize()
        self._X = None
        self._X0 = None
        self._residual = None
        self._errors = {}
        self._counter = 0
        self.initialize_unknowns_vector()
        self._make_errors()
        self.make_values()
        if self.verbose:
            # print('  *** Initialized ***')
            print(self._report)

    def reinitialization(self):
        self._preset_weights()
        self._set_dimensions()
        self._reinitialization()

    def _initialize_iteration(self):
        if self.initialize:
            self._initialization()
            self._initialize = False
            self._reinitialize = True
        self._preset_weights()
        self._set_dimensions()
        self.reinitialize_check()
        if self._reinitialize:
            self._reinitialization()
            self._reinitialize = False
        self._constraints = []
        self._constant_constraints = []
        self.__t0 = timer()

    def _pre_iteration_update(self):
        self.balance_weights()
        # self._constraints = self._constant_constraints + []

    def _post_iteration_update(self):
        self.post_iteration_update()
        self._counter += 1
        self.__counter += 1
        self.make_values()

    def _finalize(self):
        t = timer()
        self.last_iteration_time = t - self.__t0
        self.__timer += t - self.__t0
        self._make_X0()
        self._make_errors()
        if self.verbose:
            print(self._report)
        self.on_finalize()

    def _reset_counter(self):
        self._counter = 0
        self.__counter = 0
        self.__timer = 0

    def _make_X0(self):
        X = np.copy(self.X)
        self._X = None
        self.initialize_unknowns_vector()
        self._X = X

    # --------------------------------------------------------------------------
    #                               Initialization
    # --------------------------------------------------------------------------

    def _preset_weights(self):
        self.preset_weights()

    def _set_dimensions(self):
        self.set_dimensions()

    def _make_errors(self):
        self._errors = {}
        self._messages = []
        self.make_messages()
        self.make_errors()
        if self.make_residual:
            self._make_residuals()
        self._make_report()

    # --------------------------------------------------------------------------
    #                                 Overwrite
    # --------------------------------------------------------------------------

    def reinitialize_check(self):
        pass

    def on_reinitialize(self):
        pass

    def on_initialize(self):
        pass

    def on_finalize(self):
        pass

    def initialize_unknowns_vector(self):
        pass

    def preset_weights(self):
        pass

    def set_dimensions(self):
        pass

    def balance_weights(self):
        pass

    def make_errors(self):
        pass

    def make_messages(self):
        pass

    def make_values(self):
        pass

    def post_iteration_update(self):
        '''Update here objects from the unknowns'''
        pass

    # -------------------------------------------------------------------------
    #                                  Add
    # -------------------------------------------------------------------------

    def add_error(self, name, mean_error, max_error, weight=None):
        self._errors[name] = (mean_error, max_error, weight)

    def add_value(self, name, value):
        self._values[name] = value

    def add_message(self, message):
        self._messages.append(message)

    def add_weight(self, name, default=1):
        self._weights[name] = default

    def add_weights(self, weights):
        self._weights = {**self._weights, **weights}

    # --------------------------------------------------------------------------
    #                                  Get
    # --------------------------------------------------------------------------

    def get_value(self, key):
        try:
            return self._values[key]
        except KeyError:
            return None

    def get_error(self, key):
        try:
            return self._errors[key]
        except KeyError:
            return None

    def get_weight(self, key):
        return self._weights[key]

    # --------------------------------------------------------------------------
    #                                  Set
    # --------------------------------------------------------------------------

    def set_weight(self, name, value):
        if name in self._weights:
            self._weights[name] = value
        else:
            out = '<{}> weight does not exists!'.format(name)
            raise ValueError(out)

    def set_weights(self, weights):
        for name in weights:
            self.set_weight(name, weights[name])

    # --------------------------------------------------------------------------
    #                                Format
    # --------------------------------------------------------------------------

    def error_string(self, name):
        error = self.get_error(name)
        if error is None:
            return '_'
        string = '{:.4E} | {:.4E}'
        out = string.format(error[0], error[1])
        return out

    # --------------------------------------------------------------------------
    #                                 Report
    # --------------------------------------------------------------------------

    def _out_error(self, name):
        error = self.get_error(name)
        if error is None:
            error = ('-', '-', '-')
            string = '.{:>22}: {:>11} | {:>10} | {:>10} .\n'
        else:
            string = '.{:>22}: {:>11.4E} | {:>10.4E} | {:>10.4E} .\n'
        out = string.format(name, *error)
        return out

    def _format_message(self, message):
        n = 64
        mes = str(message)
        out = '. ' + mes + ' ' * (n - len(mes) - 3) + '.\n'
        return out

    def _split_string(self):
        n = 64
        out = '.' * n + '\n'
        return out

    def _separator_string(self):
        n = 64
        out = '_' * n + '\n'
        return out

    def _open_string(self):
        n = 64
        out = '-' * n + '\n'
        name = 'iteration ' + str(self.__counter)
        out += '.' + ' ' * ((n - len(name) - 1) // 2) + name
        out += ' ' * (n - (n - len(name) - 1) // 2 - len(name) - 2) + '.\n'
        out += self._split_string()
        string = '.' + ' ' * 8 + 'Relative Error:'
        string += '     mean    |    max     |   weight   .\n'
        out += string
        return out

    def _optimization_info(self):
        m = 'cumulative time: {:.3f}'.format(self.__timer)
        out = self._format_message(m)
        m = 'step: {}'.format(self._step)
        out += self._format_message(m)
        m = 'number of variables: {}'.format(self.N)
        out += self._format_message(m)
        constraints = 0
        for c in self._constraints:
            constraints += c[1]
        m = 'number of constraints: {}'.format(constraints)
        out += self._format_message(m)
        return out

    def _make_report(self):
        out = self._open_string()
        out += self._split_string()
        for error in self._errors:
            out += self._out_error(error)
        out += self._split_string()
        out += self._optimization_info()
        for message in self._messages:
            out += self._format_message(message)
        out += self._separator_string()
        self._report = out
        if self.make_residual:
            self._make_residuals_report()
        return out

    def _make_residuals_report(self):
        if self._residual is None:
            return
        out = '. Residuals\n'
        out += self._split_string()
        O = 0
        for constraint in self._constant_constraints:
            res = self._residual[O:O + constraint[1]]
            res = np.linalg.norm(res)
            out += '. {} = {:.4E}\n'.format(constraint[0], res)
            O += constraint[1]
        for constraint in self._constraints:
            res = self._residual[O:O + constraint[1]]
            res = np.linalg.norm(res)
            out += '. {} = {:.4E}\n'.format(constraint[0], res)
            O += constraint[1]
        out += self._split_string()
        self._report += out

    # -------------------------------------------------------------------------
    #                                 Save
    # -------------------------------------------------------------------------

    def save_report(self, file_name):
        try:
            name = '{}_report.txt'.format(file_name)
            txt = open(name, 'w')
            txt.write(self._report)
            txt.close()
        except:
            print('Report not available!')

    # -------------------------------------------------------------------------
    #                                Build
    # -------------------------------------------------------------------------

    def add_iterative_constraint(self, H, r, name='constraint'):
        self._H.append(H)
        self._r = np.hstack((self._r, r))
        self._constraints.append((name, H.shape[0]))

    def add_constant_constraint(self, H, r, name='constraint'):
        self._H0.append(H)
        self._r0 = np.hstack((self._r0, r))
        self._constant_constraints.append((name, H.shape[0]))

    def add_iterative_fairness(self, K, s, name='fairness'):
        self._K.append(K)
        self._s = np.hstack((self._s, s))

    def add_constant_fairness(self, K, s, name='fairness'):
        self._K0.append(K)
        self._s0 = np.hstack((self._s0, s))

    def build_iterative_constraints(self):
        pass

    def build_constant_constraints(self):
        pass

    def build_iterative_fairness(self):
        pass

    def build_constant_fairness(self):
        pass

    def build_regularizer(self):
        self._R = self.epsilon ** 2 * sparse.identity(self.N)

    # -------------------------------------------------------------------------
    #                                Build
    # -------------------------------------------------------------------------

    def _build_constant_matrices(self):
        null = np.zeros([0])
        E = sparse.coo_matrix((null, (null, null)), shape=(0, self.N))
        self._H0 = [E]
        self._K0 = [E]
        self._r0 = np.array([])
        self._s0 = np.array([])
        self.build_constant_constraints()
        self.build_constant_fairness()
        self._H0 = sparse.vstack(self._H0)
        self._K0 = sparse.vstack(self._K0)

    def _build_iterative_matrices(self):
        null = np.zeros([0])
        E = sparse.coo_matrix((null, (null, null)), shape=(0, self.N))
        self._H = [E]
        self._K = [E]
        self._r = np.array([])
        self._s = np.array([])
        self.build_iterative_constraints()
        self.build_iterative_fairness()
        self.build_regularizer()
        self._H = sparse.vstack(self._H)
        self._K = sparse.vstack(self._K)

    def _make_residuals(self):
        if self.iteration == 0:
            self._start_residual()
        H = sparse.vstack([self._H0, self._H])
        r = sparse.hstack([self._r0, self._r]).T
        r = r.toarray()
        X = sparse.csc_matrix([self.X]).transpose()
        self._residual = H.dot(X) - r

    def _start_residual(self):
        self._constant_constraints = []
        self._constraints = []
        self.balance_weights()
        self._build_constant_matrices()
        self._build_iterative_matrices()
        # self._post_iteration_update()

    # -------------------------------------------------------------------------
    #                                Solver
    # -------------------------------------------------------------------------

    def optimize(self):
        self._initialize_iteration()
        self._build_constant_matrices()
        X0 = np.array(self.X)
        if self.threshold is None:
            diff = np.linalg.norm(X0) / X0.shape[0]
        iteration = 0
        stop = False
        while iteration < self.iterations and not stop:
            self._pre_iteration_update()
            self._build_iterative_matrices()
            H = sparse.vstack([self._H0, self._H])
            r = np.array([np.hstack((self._r0, self._r))]).T
            K = sparse.vstack([self._K0, self._K])
            s = np.array([np.hstack((self._s0, self._s))]).T
            K = 1.0 / (10 ** (iteration * self.fairness_reduction)) * K
            s = 1.0 / (10 ** (iteration * self.fairness_reduction)) * s
            X = sparse.csc_matrix([self.X]).transpose()
            H = H.tocsr()
            K = K.tocsr()
            R = self._R.tocsr()
            if self.plot_matrix:
                from geolab.plotter.matrix import plot_matrix
                plot_matrix(H)
            A = spdot(H.transpose(), H) + spdot(K.transpose(), K) + R
            # A = hdot(H) + spdot(K.transpose(), K) + R
            if self.plot_matrix:
                from geolab.plotter.matrix import plot_matrix
                plot_matrix(A)
            a = spdot(H.transpose(), r) + spdot(K.transpose(), s) \
                + self.epsilon ** 2 * X
            if self.step_control and self._residual_norm is None:
                res0 = H.dot(X) - r
                res0 = np.linalg.norm(res0)
                self._residual_norm = res0
            X = spsolve(A, a)
            iteration += 1
            Xi = np.array(X)
            lam = self.step + 0
            if self.step_control:
                stop = False
                while not stop:
                    X = lam * Xi + (1 - lam) * X0
                    X = sparse.csc_matrix([X]).transpose()
                    res = H.dot(X) - r
                    res = np.linalg.norm(res)
                    if res < self._residual_norm or lam < 1e-4:
                        self._residual_norm = res
                        stop = True
                        Xi = lam * Xi + (1 - lam) * X0
                    else:
                        lam = 0.5 * lam
            elif self.step != 1:
                Xi = self.step * Xi + (1 - self.step) * X0
            if self.threshold is not None:
                diff = np.linalg.norm(X0 - Xi) / X0.shape[0]
                if diff < self.threshold:
                    stop = True
            self._step = lam
            X0 = np.array(Xi)
            self._X = np.array(Xi)
            self._post_iteration_update()
        self._finalize()
        return X0



from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, _sanity_check, _get_numpy_layout, _type_check, _create_mkl_sparse,
                                           _destroy_mkl_handle, matrix_descr, RETURN_CODES, _convert_to_csr,
                                           _check_return_value, LAYOUT_CODE_C)

import numpy as np
import ctypes as _ctypes
import scipy.sparse as _spsparse


def _sparse_qr(matrix_a, matrix_b):
    """
    Solve AX = B for X

    :param matrix_a: Sparse matrix A
    :type matrix_a: scipy.sparse.csr_matrix
    :param matrix_b: Dense matrix B
    :type matrix_b: numpy.ndarray
    :return: Dense matrix X
    :rtype: numpy.ndarray
    """

    mkl_a, dbl = _create_mkl_sparse(matrix_a)
    layout_b, ld_b = _get_numpy_layout(matrix_b)

    output_shape = matrix_a.shape[1], matrix_b.shape[1]

    if _spsparse.isspmatrix_csc(matrix_a):
        mkl_a = _convert_to_csr(mkl_a)

    # QR Reorder ##
    ret_val_r = MKL._mkl_sparse_qr_reorder(mkl_a, matrix_descr())

    # Check return
    _check_return_value(ret_val_r, "mkl_sparse_qr_reorder")

    # QR Factorize ##
    factorize_func = MKL._mkl_sparse_d_qr_factorize if dbl else MKL._mkl_sparse_s_qr_factorize

    ret_val_f = factorize_func(mkl_a, None)

    # Check return
    _check_return_value(ret_val_f, factorize_func.__name__)

    # QR Solve ##
    output_dtype = np.float64 if dbl else np.float32
    output_ctype = _ctypes.c_double if dbl else _ctypes.c_float

    output_arr = np.zeros(output_shape, dtype=output_dtype, order="C" if layout_b == LAYOUT_CODE_C else "F")
    layout_out, ld_out = _get_numpy_layout(output_arr)

    solve_func = MKL._mkl_sparse_d_qr_solve if dbl else MKL._mkl_sparse_s_qr_solve

    ret_val_s = solve_func(10,
                           mkl_a,
                           None,
                           layout_b,
                           output_shape[1],
                           output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
                           ld_out,
                           matrix_b,
                           ld_b)

    # Check return
    _check_return_value(ret_val_s, solve_func.__name__)

    _destroy_mkl_handle(mkl_a)

    return output_arr


def sparse_qr_solver(matrix_a, matrix_b, cast=False):
    """

    :param matrix_a:
    :param matrix_b:
    :param cast:
    :return:
    """

    if _spsparse.isspmatrix_csc(matrix_a) and not cast:
        raise ValueError("sparse_qr_solver only accepts CSR matrices if cast=False")
    elif not _spsparse.isspmatrix_csr(matrix_a) and not _spsparse.isspmatrix_csc(matrix_a):
        raise ValueError("sparse_qr_solver requires matrix A to be CSR or CSC sparse matrix")
    elif matrix_a.shape[0] != matrix_b.shape[0]:
        err_msg = "Bad matrix shapes for AX=B solver: A {sha} & B {shb}".format(sha=matrix_a.shape, shb=matrix_b.shape)
        raise ValueError(err_msg)
    else:
        matrix_a, matrix_b = _type_check(matrix_a, matrix_b, cast=cast)
        x_arr = _sparse_qr(matrix_a, matrix_b if matrix_b.ndim == 2 else matrix_b.reshape(-1, 1))
        return x_arr if matrix_b.ndim == 2 else x_arr.ravel()



import os
import time
import warnings
import ctypes as _ctypes
import ctypes.util as _ctypes_util

# Workaround for this stupid sklearn thing
if 'KMP_INIT_AT_FORK' in os.environ:
    _sklearn_env = os.environ['KMP_INIT_AT_FORK']
    del os.environ['KMP_INIT_AT_FORK']
else:
    _sklearn_env = None

# Load mkl_spblas through the libmkl_rt common interface
_libmkl = None
try:
    _so_file = _ctypes_util.find_library('mkl_rt')
    if _so_file is None:
        # For some reason, find_library is not checking LD_LIBRARY_PATH
        # If the ctypes.util approach doesn't work, try this (crude) approach

        # Check each of these library types
        for so_file in ["libmkl_rt.so", "libmkl_rt.dylib", "mkl_rt.dll"]:
            try:
                _libmkl = _ctypes.cdll.LoadLibrary(so_file)
                break
            except (OSError, ImportError) as err:
                pass

        if _libmkl is None:
            raise ImportError("mkl_rt not found.")
    else:
        _libmkl = _ctypes.cdll.LoadLibrary(_so_file)
except (OSError, ImportError) as err:
    _ierr_msg = "Unable to load the MKL libraries through libmkl_rt. Try setting $LD_LIBRARY_PATH. " + str(err)
    raise ImportError(_ierr_msg)

# Use mkl-service to check version if it's installed
# Since it's not on PyPi I don't want to make this an actual package dependency
# So without it just create mock functions and don't do version checking
try:
    from mkl import get_version, get_version_string
except ImportError:
    def get_version():
        return None


    def get_version_string():
        return None

if get_version() is not None and get_version()["MajorVersion"] < 2020:
    _verr_msg = "Loaded version of MKL is out of date: {v}".format(v=get_version_string())
    warnings.warn(_verr_msg)

import numpy as np
import scipy.sparse as _spsparse
from numpy.ctypeslib import ndpointer, as_array

NUMPY_FLOAT_DTYPES = [np.float32, np.float64]


class MKL:
    """ This class holds shared object references to C functions with arg and returntypes that can be adjusted"""

    MKL_INT = None
    MKL_INT_NUMPY = None
    MKL_DEBUG = False

    # Import function for creating a MKL CSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-csr
    _mkl_sparse_d_create_csr = _libmkl.mkl_sparse_d_create_csr

    # Import function for creating a MKL CSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-csr
    _mkl_sparse_s_create_csr = _libmkl.mkl_sparse_s_create_csr

    # Import function for creating a MKL CSC object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-csc
    _mkl_sparse_d_create_csc = _libmkl.mkl_sparse_d_create_csc

    # Import function for creating a MKL CSC object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-csc
    _mkl_sparse_s_create_csc = _libmkl.mkl_sparse_s_create_csc

    # Import function for creating a MKL BSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-bsr
    _mkl_sparse_d_create_bsr = _libmkl.mkl_sparse_d_create_bsr

    # Import function for creating a MKL BSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-create-bsr
    _mkl_sparse_s_create_bsr = _libmkl.mkl_sparse_s_create_bsr

    # Export function for exporting a MKL CSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-csr
    _mkl_sparse_d_export_csr = _libmkl.mkl_sparse_d_export_csr

    # Export function for exporting a MKL CSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-csr
    _mkl_sparse_s_export_csr = _libmkl.mkl_sparse_s_export_csr

    # Export function for exporting a MKL CSC object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-csc
    _mkl_sparse_d_export_csc = _libmkl.mkl_sparse_d_export_csc

    # Export function for exporting a MKL CSC object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-csc
    _mkl_sparse_s_export_csc = _libmkl.mkl_sparse_s_export_csc

    # Export function for exporting a MKL BSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-bsr
    _mkl_sparse_d_export_bsr = _libmkl.mkl_sparse_d_export_bsr

    # Export function for exporting a MKL BSR object
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-export-bsr
    _mkl_sparse_s_export_bsr = _libmkl.mkl_sparse_s_export_bsr

    # Import function for matmul
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-spmm
    _mkl_sparse_spmm = _libmkl.mkl_sparse_spmm

    # Import function for cleaning up MKL objects
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-destroy
    _mkl_sparse_destroy = _libmkl.mkl_sparse_destroy

    # Import function for ordering MKL objects
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-order
    _mkl_sparse_order = _libmkl.mkl_sparse_order

    # Import function for coverting to CSR
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-convert-csr
    _mkl_sparse_convert_csr = _libmkl.mkl_sparse_convert_csr

    # Import function for matmul single dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-spmm
    _mkl_sparse_s_spmmd = _libmkl.mkl_sparse_s_spmmd

    # Import function for matmul double dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-spmm
    _mkl_sparse_d_spmmd = _libmkl.mkl_sparse_d_spmmd

    # Import function for matmul single sparse*dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-mm
    _mkl_sparse_s_mm = _libmkl.mkl_sparse_s_mm

    # Import function for matmul double sparse*dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-mm
    _mkl_sparse_d_mm = _libmkl.mkl_sparse_d_mm

    # Import function for matmul single dense*dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-cblas-gemm
    _cblas_sgemm = _libmkl.cblas_sgemm

    # Import function for matmul double dense*dense
    # https://software.intel.com/en-us/mkl-developer-reference-c-cblas-gemm
    _cblas_dgemm = _libmkl.cblas_dgemm

    # Import function for matrix * vector
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-mv
    _mkl_sparse_s_mv = _libmkl.mkl_sparse_s_mv

    # Import function for matrix * vector
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-mv
    _mkl_sparse_d_mv = _libmkl.mkl_sparse_d_mv

    # Import function for sparse gram matrix
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-syrk
    _mkl_sparse_syrk = _libmkl.mkl_sparse_syrk

    # Import function for dense single gram matrix from sparse
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-syrkd
    _mkl_sparse_s_syrkd = _libmkl.mkl_sparse_s_syrkd

    # Import function for dense double gram matrix from sparse
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-syrkd
    _mkl_sparse_d_syrkd = _libmkl.mkl_sparse_d_syrkd

    # Import function for dense single gram matrix
    # https://software.intel.com/en-us/mkl-developer-reference-c-cblas-syrk
    _cblas_ssyrk = _libmkl.cblas_ssyrk

    # Import function for dense double gram matrix
    # https://software.intel.com/en-us/mkl-developer-reference-c-cblas-syrk
    _cblas_dsyrk = _libmkl.cblas_dsyrk

    # Import function for QR solver - reorder
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-qr-reorder
    _mkl_sparse_qr_reorder = _libmkl.mkl_sparse_qr_reorder

    # Import function for QR solver - factorize
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-qr-factorize
    _mkl_sparse_d_qr_factorize = _libmkl.mkl_sparse_d_qr_factorize

    # Import function for QR solver - factorize
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-qr-factorize
    _mkl_sparse_s_qr_factorize = _libmkl.mkl_sparse_s_qr_factorize

    # Import function for QR solver - solve
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-qr-solve
    _mkl_sparse_d_qr_solve = _libmkl.mkl_sparse_d_qr_solve

    # Import function for QR solver - solve
    # https://software.intel.com/en-us/mkl-developer-reference-c-mkl-sparse-qr-solve
    _mkl_sparse_s_qr_solve = _libmkl.mkl_sparse_s_qr_solve

    @classmethod
    def _set_int_type(cls, c_type, np_type):
        cls.MKL_INT = c_type
        cls.MKL_INT_NUMPY = np_type

        cls._mkl_sparse_d_create_csr.argtypes = cls._mkl_sparse_create_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_create_csr.restypes = _ctypes.c_int

        cls._mkl_sparse_s_create_csr.argtypes = cls._mkl_sparse_create_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_create_csr.restypes = _ctypes.c_int

        cls._mkl_sparse_d_create_csc.argtypes = cls._mkl_sparse_create_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_create_csc.restypes = _ctypes.c_int

        cls._mkl_sparse_s_create_csc.argtypes = cls._mkl_sparse_create_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_create_csc.restypes = _ctypes.c_int

        cls._mkl_sparse_d_create_bsr.argtypes = cls._mkl_sparse_create_bsr_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_create_bsr.restypes = _ctypes.c_int

        cls._mkl_sparse_s_create_bsr.argtypes = cls._mkl_sparse_create_bsr_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_create_bsr.restypes = _ctypes.c_int

        cls._mkl_sparse_d_export_csr.argtypes = cls._mkl_sparse_export_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_export_csr.restypes = _ctypes.c_int

        cls._mkl_sparse_s_export_csr.argtypes = cls._mkl_sparse_export_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_export_csr.restypes = _ctypes.c_int

        cls._mkl_sparse_d_export_csc.argtypes = cls._mkl_sparse_export_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_export_csc.restypes = _ctypes.c_int

        cls._mkl_sparse_s_export_bsr.argtypes = cls._mkl_sparse_export_bsr_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_export_bsr.restypes = _ctypes.c_int

        cls._mkl_sparse_s_export_bsr.argtypes = cls._mkl_sparse_export_bsr_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_export_bsr.restypes = _ctypes.c_int

        cls._mkl_sparse_spmm.argtypes = [_ctypes.c_int,
                                         sparse_matrix_t,
                                         sparse_matrix_t,
                                         _ctypes.POINTER(sparse_matrix_t)]
        cls._mkl_sparse_spmm.restypes = _ctypes.c_int

        cls._mkl_sparse_s_spmmd.argtypes = cls._mkl_sparse_spmmd_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_spmmd.restypes = _ctypes.c_int

        cls._mkl_sparse_d_spmmd.argtypes = cls._mkl_sparse_spmmd_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_spmmd.restypes = _ctypes.c_int

        cls._mkl_sparse_s_mm.argtypes = cls._mkl_sparse_mm_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_mm.restypes = _ctypes.c_int

        cls._mkl_sparse_d_mm.argtypes = cls._mkl_sparse_mm_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_mm.restypes = _ctypes.c_int

        cls._cblas_sgemm.argtypes = cls._cblas_gemm_argtypes(_ctypes.c_float)
        cls._cblas_sgemm.restypes = None

        cls._cblas_dgemm.argtypes = cls._cblas_gemm_argtypes(_ctypes.c_double)
        cls._cblas_dgemm.restypes = None

        cls._mkl_sparse_destroy.argtypes = [sparse_matrix_t]
        cls._mkl_sparse_destroy.restypes = _ctypes.c_int

        cls._mkl_sparse_order.argtypes = [sparse_matrix_t]
        cls._mkl_sparse_order.restypes = _ctypes.c_int

        cls._mkl_sparse_s_mv.argtypes = cls._mkl_sparse_mv_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_mv.restypes = _ctypes.c_int

        cls._mkl_sparse_d_mv.argtypes = cls._mkl_sparse_mv_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_mv.restypes = _ctypes.c_int

        cls._mkl_sparse_syrk.argtypes = [_ctypes.c_int,
                                         sparse_matrix_t,
                                         _ctypes.POINTER(sparse_matrix_t)]
        cls._mkl_sparse_syrk.restypes = _ctypes.c_int

        cls._mkl_sparse_s_syrkd.argtypes = cls._mkl_sparse_syrkd_argtypes(_ctypes.c_float)
        cls._mkl_sparse_s_syrkd.restypes = _ctypes.c_int

        cls._mkl_sparse_d_syrkd.argtypes = cls._mkl_sparse_syrkd_argtypes(_ctypes.c_double)
        cls._mkl_sparse_d_syrkd.restypes = _ctypes.c_int

        cls._cblas_ssyrk.argtypes = cls._cblas_syrk_argtypes(_ctypes.c_float)
        cls._cblas_ssyrk.restypes = None

        cls._cblas_dsyrk.argtypes = cls._cblas_syrk_argtypes(_ctypes.c_double)
        cls._cblas_dsyrk.restypes = None

        cls._mkl_sparse_qr_reorder.argtypes = [sparse_matrix_t, matrix_descr]
        cls._mkl_sparse_qr_reorder.restypes = _ctypes.c_int

        cls._mkl_sparse_d_qr_factorize.argtypes = [sparse_matrix_t, _ctypes.POINTER(_ctypes.c_double)]
        cls._mkl_sparse_d_qr_factorize.restypes = _ctypes.c_int

        cls._mkl_sparse_s_qr_factorize.argtypes = [sparse_matrix_t, _ctypes.POINTER(_ctypes.c_float)]
        cls._mkl_sparse_s_qr_factorize.restypes = _ctypes.c_int

        cls._mkl_sparse_d_qr_solve.argtypes = cls._mkl_sparse_qr_solve(_ctypes.c_double)
        cls._mkl_sparse_d_qr_solve.restypes = _ctypes.c_int

        cls._mkl_sparse_s_qr_solve.argtypes = cls._mkl_sparse_qr_solve(_ctypes.c_float)
        cls._mkl_sparse_s_qr_solve.restypes = _ctypes.c_int

    def __init__(self):
        raise NotImplementedError("This class is not intended to be instanced")

    """ The following methods return the argtype lists for each MKL function that has s and d variants"""

    @staticmethod
    def _mkl_sparse_create_argtypes(prec_type):
        return [_ctypes.POINTER(sparse_matrix_t),
                _ctypes.c_int,
                MKL.MKL_INT,
                MKL.MKL_INT,
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=prec_type, ndim=1, flags='C_CONTIGUOUS')]

    @staticmethod
    def _mkl_sparse_create_bsr_argtypes(prec_type):
        return [_ctypes.POINTER(sparse_matrix_t),
                _ctypes.c_int,
                _ctypes.c_int,
                MKL.MKL_INT,
                MKL.MKL_INT,
                MKL.MKL_INT,
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=MKL.MKL_INT, ndim=1, flags='C_CONTIGUOUS'),
                ndpointer(dtype=prec_type, ndim=3, flags='C_CONTIGUOUS')]

    @staticmethod
    def _mkl_sparse_export_argtypes(prec_type):
        return [sparse_matrix_t,
                _ctypes.POINTER(_ctypes.c_int),
                _ctypes.POINTER(MKL.MKL_INT),
                _ctypes.POINTER(MKL.MKL_INT),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(prec_type))]

    @staticmethod
    def _mkl_sparse_export_bsr_argtypes(prec_type):
        return [sparse_matrix_t,
                _ctypes.POINTER(_ctypes.c_int),
                _ctypes.POINTER(_ctypes.c_int),
                _ctypes.POINTER(MKL.MKL_INT),
                _ctypes.POINTER(MKL.MKL_INT),
                _ctypes.POINTER(MKL.MKL_INT),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(MKL.MKL_INT)),
                _ctypes.POINTER(_ctypes.POINTER(prec_type))]

    @staticmethod
    def _cblas_gemm_argtypes(prec_type):
        return [_ctypes.c_int,
                _ctypes.c_int,
                _ctypes.c_int,
                MKL.MKL_INT,
                MKL.MKL_INT,
                MKL.MKL_INT,
                prec_type,
                ndpointer(dtype=prec_type, ndim=2),
                MKL.MKL_INT,
                ndpointer(dtype=prec_type, ndim=2),
                MKL.MKL_INT,
                prec_type,
                _ctypes.POINTER(prec_type),
                MKL.MKL_INT]

    @staticmethod
    def _mkl_sparse_spmmd_argtypes(prec_type):
        return [_ctypes.c_int,
                sparse_matrix_t,
                sparse_matrix_t,
                _ctypes.c_int,
                _ctypes.POINTER(prec_type), MKL.MKL_INT]

    @staticmethod
    def _mkl_sparse_mm_argtypes(prec_type):
        return [_ctypes.c_int,
                prec_type,
                sparse_matrix_t,
                matrix_descr,
                _ctypes.c_int,
                ndpointer(dtype=prec_type, ndim=2),
                MKL.MKL_INT,
                MKL.MKL_INT,
                prec_type,
                _ctypes.POINTER(prec_type),
                MKL.MKL_INT]

    @staticmethod
    def _mkl_sparse_mv_argtypes(prec_type):
        return [_ctypes.c_int,
                prec_type,
                sparse_matrix_t,
                matrix_descr,
                ndpointer(dtype=prec_type, ndim=1),
                prec_type,
                _ctypes.POINTER(prec_type)]

    @staticmethod
    def _mkl_sparse_syrkd_argtypes(prec_type):
        return [_ctypes.c_int,
                sparse_matrix_t,
                prec_type,
                prec_type,
                _ctypes.POINTER(prec_type),
                _ctypes.c_int,
                MKL.MKL_INT]

    @staticmethod
    def _cblas_syrk_argtypes(prec_type):
        return [_ctypes.c_int,
                _ctypes.c_int,
                _ctypes.c_int,
                MKL.MKL_INT,
                MKL.MKL_INT,
                prec_type,
                ndpointer(dtype=prec_type, ndim=2),
                MKL.MKL_INT,
                prec_type,
                _ctypes.POINTER(prec_type),
                MKL.MKL_INT]

    @staticmethod
    def _mkl_sparse_qr_solve(prec_type):
        return [_ctypes.c_int,
                sparse_matrix_t,
                _ctypes.POINTER(prec_type),
                _ctypes.c_int,
                MKL.MKL_INT,
                _ctypes.POINTER(prec_type),
                MKL.MKL_INT,
                ndpointer(dtype=prec_type, ndim=2),
                MKL.MKL_INT]


# Construct opaque struct & type
class _sparse_matrix(_ctypes.Structure):
    pass


sparse_matrix_t = _ctypes.POINTER(_sparse_matrix)


# Matrix description struct
class matrix_descr(_ctypes.Structure):
    _fields_ = [("sparse_matrix_type_t", _ctypes.c_int),
                ("sparse_fill_mode_t", _ctypes.c_int),
                ("sparse_diag_type_t", _ctypes.c_int)]

    def __init__(self, sparse_matrix_type_t=20, sparse_fill_mode_t=0, sparse_diag_type_t=0):
        super(matrix_descr, self).__init__(sparse_matrix_type_t, sparse_fill_mode_t, sparse_diag_type_t)


# Define standard return codes
RETURN_CODES = {0: "SPARSE_STATUS_SUCCESS",
                1: "SPARSE_STATUS_NOT_INITIALIZED",
                2: "SPARSE_STATUS_ALLOC_FAILED",
                3: "SPARSE_STATUS_INVALID_VALUE",
                4: "SPARSE_STATUS_EXECUTION_FAILED",
                5: "SPARSE_STATUS_INTERNAL_ERROR",
                6: "SPARSE_STATUS_NOT_SUPPORTED"}

# Define order codes
LAYOUT_CODE_C = 101
LAYOUT_CODE_F = 102

# Define transpose codes
SPARSE_OPERATION_NON_TRANSPOSE = 10
SPARSE_OPERATION_TRANSPOSE = 11

# Define index codes
SPARSE_INDEX_BASE_ZERO = 0
SPARSE_INDEX_BASE_ONE = 1

# ILP64 message
ILP64_MSG = " Try changing MKL to int64 with the environment variable MKL_INTERFACE_LAYER=ILP64"


def set_debug_mode(debug_bool):
    """
    Activate or deactivate debug mode

    :param debug_bool: True to be printy. False to be quiet.
    :type debug_bool: bool
    """

    MKL.MKL_DEBUG = debug_bool


def print_mkl_debug():
    """
    Print the MKL interface status if debug mode is on
    """

    if not MKL.MKL_DEBUG:
        return

    if get_version_string() is None:
        print("mkl-service must be installed to get full debug messaging")
    else:
        print(get_version_string())

    print("MKL linked: {fn}".format(fn=_libmkl._name))
    print("MKL interface {np} | {c}".format(np=MKL.MKL_INT_NUMPY, c=MKL.MKL_INT))
    print("Set int32 interface with env MKL_INTERFACE_LAYER=LP64")
    print("Set int64 interface with env MKL_INTERFACE_LAYER=ILP64")


def debug_print(msg):
    """
    Print a message if debug mode is on
    :param msg: Message
    :type msg: str
    """

    if not MKL.MKL_DEBUG:
        return
    else:
        print(msg)


def debug_timer(msg=None, old_time=None):
    """
    Print a message with timing information if debug mode is on
    :param msg: Message
    :type msg: str
    :param old_time: Time to calculate difference for
    :type old_time: float
    """

    if not MKL.MKL_DEBUG:
        return

    t0 = time.time()

    if old_time is not None and msg is not None:
        print(msg + ": {0:.6f} seconds".format(t0 - old_time))

    return t0


def _check_scipy_index_typing(sparse_matrix):
    """
    Ensure that the sparse matrix indicies are in the correct integer type

    :param sparse_matrix: Scipy matrix in CSC or CSR format
    :type sparse_matrix: scipy.sparse.spmatrix
    """

    int_max = np.iinfo(MKL.MKL_INT_NUMPY).max
    if (sparse_matrix.nnz > int_max) or (max(sparse_matrix.shape) > int_max):
        msg = "MKL interface is {t} and cannot hold matrix {m}\n".format(m=repr(sparse_matrix), t=MKL.MKL_INT_NUMPY)
        msg += "Try changing MKL to int64 with the environment variable MKL_INTERFACE_LAYER=ILP64"
        raise ValueError(msg)

    # Cast indexes to MKL_INT type
    if sparse_matrix.indptr.dtype != MKL.MKL_INT_NUMPY:
        sparse_matrix.indptr = sparse_matrix.indptr.astype(MKL.MKL_INT_NUMPY)
    if sparse_matrix.indices.dtype != MKL.MKL_INT_NUMPY:
        sparse_matrix.indices = sparse_matrix.indices.astype(MKL.MKL_INT_NUMPY)


def _get_numpy_layout(numpy_arr, second_arr=None):
    """
    Get the array layout code for a dense array in C or F order.
    Raises a ValueError if the array is not contiguous.

    :param numpy_arr: Numpy dense array
    :type numpy_arr: np.ndarray
    :param second_arr: Numpy dense array; if numpy_arr is 1d (and therefore both C and F order), use this order
    :type second_arr: np.ndarray, None
    :return: The layout code for MKL and the leading dimension
    :rtype: int, int
    """

    # Return the second array order if the first is ambiguous
    if numpy_arr.flags.c_contiguous and numpy_arr.flags.f_contiguous and second_arr is not None:
        if second_arr.flags.c_contiguous:
            return LAYOUT_CODE_C, numpy_arr.shape[1]
        elif second_arr.flags.f_contiguous:
            return LAYOUT_CODE_F, numpy_arr.shape[0]

    # Return the first order array otherwise
    elif numpy_arr.flags.c_contiguous:
        return LAYOUT_CODE_C, numpy_arr.shape[1]
    elif numpy_arr.flags.f_contiguous:
        return LAYOUT_CODE_F, numpy_arr.shape[0]
    elif not numpy_arr.flags.contiguous:
        raise ValueError("Array is not contiguous")
    else:
        raise ValueError("Array layout check has failed for unknown reason")


def _create_mkl_sparse(matrix):
    """
    Create MKL internal representation

    :param matrix: Sparse data in CSR or CSC format
    :type matrix: scipy.sparse.spmatrix

    :return ref, double_precision: Handle for the MKL internal representation and boolean for double precision
    :rtype: sparse_matrix_t, float
    """

    double_precision = _is_double(matrix)

    # Figure out which matrix creation function to use
    if _spsparse.isspmatrix_csr(matrix):
        _check_scipy_index_typing(matrix)
        assert matrix.data.shape[0] == matrix.indices.shape[0]
        assert matrix.indptr.shape[0] == matrix.shape[0] + 1
        handle_func = MKL._mkl_sparse_d_create_csr if double_precision else MKL._mkl_sparse_s_create_csr

    elif _spsparse.isspmatrix_csc(matrix):
        _check_scipy_index_typing(matrix)
        assert matrix.data.shape[0] == matrix.indices.shape[0]
        assert matrix.indptr.shape[0] == matrix.shape[1] + 1
        handle_func = MKL._mkl_sparse_d_create_csc if double_precision else MKL._mkl_sparse_s_create_csc

    elif _spsparse.isspmatrix_bsr(matrix):
        _check_scipy_index_typing(matrix)
        return _create_mkl_sparse_bsr(matrix), double_precision

    else:
        raise ValueError("Matrix is not CSC, CSR, or BSR")

    return _pass_mkl_handle_csr_csc(matrix, handle_func), double_precision


def _pass_mkl_handle_csr_csc(data, handle_func):
    """
    Create MKL internal representation for CSR or CSC matrix

    :param data: Sparse data
    :type data: scipy.sparse.spmatrix
    :return ref: Handle for the MKL internal representation
    :rtype: sparse_matrix_t
    """

    # Create a pointer for the output matrix
    ref = sparse_matrix_t()

    # Load into a MKL data structure and check return
    ret_val = handle_func(_ctypes.byref(ref),
                          _ctypes.c_int(SPARSE_INDEX_BASE_ZERO),
                          MKL.MKL_INT(data.shape[0]),
                          MKL.MKL_INT(data.shape[1]),
                          data.indptr[0:-1],
                          data.indptr[1:],
                          data.indices,
                          data.data)

    # Check return
    _check_return_value(ret_val, handle_func.__name__)

    return ref


def _create_mkl_sparse_bsr(matrix):
    """
    Create MKL internal representation for BSR matrix

    :param matrix: Sparse data
    :type matrix: scipy.sparse.bsr_matrix
    :return ref: Handle for the MKL internal representation
    :rtype: sparse_matrix_t
    """

    double_precision = _is_double(matrix)
    handle_func = MKL._mkl_sparse_d_create_bsr if double_precision else MKL._mkl_sparse_s_create_bsr

    # Get the blocksize and check that the blocks are square
    _blocksize = matrix.blocksize[0]

    if _blocksize != matrix.blocksize[1]:
        _err = "MKL BSR representation requires square blocks; {n} blocks provided".format(n=matrix.blocksize)
        raise ValueError(_err)

    if (matrix.shape[0] % _blocksize != 0) or (matrix.shape[1] % _blocksize != 0):
        _err = "BSR blocks {n} do not align with dims {m}".format(n=matrix.blocksize, m=matrix.shape)
        raise ValueError(_err)

    _block_rows = int(matrix.shape[0] / _blocksize)
    _block_cols = int(matrix.shape[1] / _blocksize)

    # Get the data block array structure
    _layout, _ = _get_numpy_layout(matrix.data)

    # Create a pointer for the output matrix
    ref = sparse_matrix_t()

    # Load into a MKL data structure and check return
    ret_val = handle_func(_ctypes.byref(ref),
                          _ctypes.c_int(SPARSE_INDEX_BASE_ZERO),
                          _ctypes.c_int(_layout),
                          MKL.MKL_INT(_block_rows),
                          MKL.MKL_INT(_block_cols),
                          MKL.MKL_INT(_blocksize),
                          matrix.indptr[0:-1],
                          matrix.indptr[1:],
                          matrix.indices,
                          matrix.data)

    # Check return
    _check_return_value(ret_val, handle_func.__name__)

    return ref


def _export_mkl(csr_mkl_handle, double_precision, output_type="csr"):
    """
    Export a MKL sparse handle of CSR or CSC type

    :param csr_mkl_handle: Handle for the MKL internal representation
    :type csr_mkl_handle: sparse_matrix_t
    :param double_precision: Use float64 if True, float32 if False. This MUST match the underlying float type - this
        defines a memory view, it does not cast.
    :type double_precision: bool
    :param output_type: The structure of the MKL handle (and therefore the type of scipy sparse to create)
    :type output_type: str

    :return: Sparse matrix in scipy format
    :rtype: scipy.spmatrix
    """

    output_type = output_type.lower()

    if output_type == "csr":
        out_func = MKL._mkl_sparse_d_export_csr if double_precision else MKL._mkl_sparse_s_export_csr
        sp_matrix_constructor = _spsparse.csr_matrix
    elif output_type == "csc":
        out_func = MKL._mkl_sparse_d_export_csc if double_precision else MKL._mkl_sparse_s_export_csc
        sp_matrix_constructor = _spsparse.csc_matrix
    elif output_type == "bsr":
        return _export_mkl_sparse_bsr(csr_mkl_handle, double_precision)
    else:
        raise ValueError("Only CSR, CSC, and BSR output types are supported")

    # Allocate for output
    ordering, nrows, ncols, indptrb, indptren, indices, data = _allocate_for_export(double_precision)
    final_dtype = np.float64 if double_precision else np.float32

    ret_val = out_func(csr_mkl_handle,
                       _ctypes.byref(ordering),
                       _ctypes.byref(nrows),
                       _ctypes.byref(ncols),
                       _ctypes.byref(indptrb),
                       _ctypes.byref(indptren),
                       _ctypes.byref(indices),
                       _ctypes.byref(data))

    # Check return
    _check_return_value(ret_val, out_func.__name__)

    # Check ordering
    if ordering.value != 0:
        raise ValueError("1-indexing (F-style) is not supported")

    # Get matrix dims
    ncols, nrows = ncols.value, nrows.value

    # If any axis is 0 return an empty matrix
    if nrows == 0 or ncols == 0:
        return sp_matrix_constructor((nrows, ncols), dtype=final_dtype)

    # Get the index dimension
    index_dim = nrows if output_type == "csr" else ncols

    # Construct a numpy array and add 0 to first position for scipy.sparse's 3-array indexing
    indptrb = as_array(indptrb, shape=(index_dim,))
    indptren = as_array(indptren, shape=(index_dim,))

    indptren = np.insert(indptren, 0, indptrb[0])
    nnz = indptren[-1] - indptrb[0]

    # If there are no non-zeros, return an empty matrix
    # If the number of non-zeros is insane, raise a ValueError
    if nnz == 0:
        return sp_matrix_constructor((nrows, ncols), dtype=final_dtype)
    elif nnz < 0 or nnz > ncols * nrows:
        raise ValueError("Matrix ({m} x {n}) is attempting to index {z} elements".format(m=nrows, n=ncols, z=nnz))

    # Construct numpy arrays from data pointer and from indicies pointer
    data = np.array(as_array(data, shape=(nnz,)), copy=True)
    indices = np.array(as_array(indices, shape=(nnz,)), copy=True)

    # Pack and return the matrix
    return sp_matrix_constructor((data, indices, indptren), shape=(nrows, ncols))


def _export_mkl_sparse_bsr(bsr_mkl_handle, double_precision):
    """
    Export a BSR matrix from MKL's internal representation to scipy

    :param bsr_mkl_handle: MKL internal representation
    :type bsr_mkl_handle: sparse_matrix_t
    :param double_precision: Use float64 if True, float32 if False. This MUST match the underlying float type - this
        defines a memory view, it does not cast.
    :type double_precision: bool
    :return: Sparse BSR matrix
    :rtype:
    """

    # Allocate for output
    ordering, nrows, ncols, indptrb, indptren, indices, data = _allocate_for_export(double_precision)
    block_layout = _ctypes.c_int()
    block_size = MKL.MKL_INT()

    # Set output
    out_func = MKL._mkl_sparse_d_export_bsr if double_precision else MKL._mkl_sparse_s_export_bsr
    final_dtype = np.float64 if double_precision else np.float32

    ret_val = out_func(bsr_mkl_handle,
                       _ctypes.byref(ordering),
                       _ctypes.byref(block_layout),
                       _ctypes.byref(nrows),
                       _ctypes.byref(ncols),
                       _ctypes.byref(block_size),
                       _ctypes.byref(indptrb),
                       _ctypes.byref(indptren),
                       _ctypes.byref(indices),
                       _ctypes.byref(data))

    # Check return
    _check_return_value(ret_val, out_func.__name__)

    # Get matrix dims
    ncols, nrows, block_size = ncols.value, nrows.value, block_size.value
    index_dim, block_dims = nrows, (block_size, block_size)
    ncols, nrows = ncols * block_size, nrows * block_size

    # If any axis is 0 return an empty matrix
    if nrows == 0 or ncols == 0:
        return _spsparse.bsr_matrix((nrows, ncols), dtype=final_dtype, blocksize=block_dims)

    ordering = "F" if ordering.value == LAYOUT_CODE_F else "C"

    # Construct a numpy array and add 0 to first position for scipy.sparse's 3-array indexing
    indptrb = as_array(indptrb, shape=(index_dim,))
    indptren = as_array(indptren, shape=(index_dim,))

    indptren = np.insert(indptren, 0, indptrb[0])

    nnz_blocks = (indptren[-1] - indptrb[0])

    # If there's no non-zero data, return an empty matrix
    if nnz_blocks == 0:
        return _spsparse.bsr_matrix((nrows, ncols), dtype=final_dtype, blocksize=block_dims)
    elif nnz_blocks < 0 or (nnz_blocks * (block_size ** 2)) > ncols * nrows:
        nnz = nnz_blocks * (block_size ** 2)
        _err = "Matrix ({m} x {n}) is attempting to index {z} elements as {b} {bs} blocks"
        _err = _err.format(m=nrows, n=ncols, z=nnz, b=nnz_blocks, bs=(block_size, block_size))
        raise ValueError(_err)

    data = np.array(as_array(data, shape=(nnz_blocks, block_size, block_size)), copy=True, order=ordering,
                    dtype=final_dtype)
    indices = np.array(as_array(indices, shape=(nnz_blocks,)), copy=True)

    return _spsparse.bsr_matrix((data, indices, indptren), shape=(nrows, ncols), blocksize=block_dims)


def _allocate_for_export(double_precision):
    """
    Get pointers for output from MKL internal representation
    :param double_precision: Allocate an output pointer of doubles
    :type double_precision: bool
    :return: ordering, nrows, ncols, indptrb, indptren, indices, data
    :rtype: c_int, MKL_INT, MKL_INT, MKL_INT*, MKL_INT*, MKL_INT*, c_float|c_double*
    """
    # Create the pointers for the output data
    indptrb = _ctypes.POINTER(MKL.MKL_INT)()
    indptren = _ctypes.POINTER(MKL.MKL_INT)()
    indices = _ctypes.POINTER(MKL.MKL_INT)()

    ordering = _ctypes.c_int()
    nrows = MKL.MKL_INT()
    ncols = MKL.MKL_INT()

    data = _ctypes.POINTER(_ctypes.c_double)() if double_precision else _ctypes.POINTER(_ctypes.c_float)()

    return ordering, nrows, ncols, indptrb, indptren, indices, data


def _check_return_value(ret_val, func_name):
    """
    Check the return value from a sparse function

    :param ret_val:
    :param func_name:
    :return:
    """

    if ret_val != 0:
        err_msg = "{fn} returned {v} ({e})".format(fn=func_name, v=ret_val, e=RETURN_CODES[ret_val])
        if ret_val == 2:
            err_msg += "; " + ILP64_MSG
        raise ValueError(err_msg)
    elif MKL.MKL_DEBUG:
        print("{fn} returned {v} ({e})".format(fn=func_name, v=ret_val, e=RETURN_CODES[ret_val]))
    else:
        return


def _destroy_mkl_handle(ref_handle):
    """
    Deallocate a MKL sparse handle

    :param ref_handle:
    :type ref_handle: sparse_matrix_t
    """

    ret_val = MKL._mkl_sparse_destroy(ref_handle)
    _check_return_value(ret_val, "mkl_sparse_destroy")


def _order_mkl_handle(ref_handle):
    """
    Reorder indexes in a MKL sparse handle

    :param ref_handle:
    :type ref_handle: sparse_matrix_t
    """

    ret_val = MKL._mkl_sparse_order(ref_handle)
    _check_return_value(ret_val, "mkl_sparse_order")


def _convert_to_csr(ref_handle, destroy_original=False):
    """
    Convert a MKL sparse handle to CSR format

    :param ref_handle:
    :type ref_handle: sparse_matrix_t
    :return:
    """

    csr_ref = sparse_matrix_t()
    ret_val = MKL._mkl_sparse_convert_csr(ref_handle, _ctypes.c_int(10), _ctypes.byref(csr_ref))

    try:
        _check_return_value(ret_val, "mkl_sparse_convert_csr")
    except ValueError:
        try:
            _destroy_mkl_handle(csr_ref)
        except ValueError:
            pass

        raise

    if destroy_original:
        _destroy_mkl_handle(ref_handle)

    return csr_ref


def _sanity_check(matrix_a, matrix_b, allow_vector=False):
    """
    Check matrix dimensions
    :param matrix_a: sp.sparse or numpy array
    :param matrix_b: sp.sparse or numpy array
    """

    a_2d, b_2d = matrix_a.ndim == 2, matrix_b.ndim == 2
    a_vec, b_vec = _is_dense_vector(matrix_a), _is_dense_vector(matrix_b)

    # Check to make sure that both matrices are 2-d
    if not allow_vector and (not a_2d or not b_2d):
        err_msg = "Matrices must be 2d: {m1} * {m2} is not valid".format(m1=matrix_a.shape, m2=matrix_b.shape)
        raise ValueError(err_msg)

    invalid_ndims = not (a_2d or a_vec) or not (b_2d or b_vec)
    invalid_align = (matrix_a.shape[1] if not matrix_a.ndim == 1 else matrix_a.shape[0]) != matrix_b.shape[0]

    # Check to make sure that this multiplication can work
    if invalid_align or invalid_ndims:
        err_msg = "Matrix alignment error: {m1} * {m2} is not valid".format(m1=matrix_a.shape, m2=matrix_b.shape)
        raise ValueError(err_msg)


def _cast_to_float64(matrix):
    """ Make a copy of the array as double precision floats or return the reference if it already is"""
    return matrix.astype(np.float64) if matrix.dtype != np.float64 else matrix


def _type_check(matrix_a, matrix_b=None, cast=False):
    """
    Make sure that both matrices are single precision floats or both are double precision floats
    If not, convert to double precision floats if cast is True, or raise an error if cast is False
    """

    if matrix_b is None and matrix_a.dtype in NUMPY_FLOAT_DTYPES:
        return matrix_a
    elif matrix_b is None and cast:
        return _cast_to_float64(matrix_a)
    elif matrix_b is None:
        err_msg = "Matrix data type must be float32 or float64; {a} provided".format(a=matrix_a.dtype)
        raise ValueError(err_msg)

    # Check dtypes
    if matrix_a.dtype == np.float32 and matrix_b.dtype == np.float32:
        return matrix_a, matrix_b

    elif matrix_a.dtype == np.float64 and matrix_b.dtype == np.float64:
        return matrix_a, matrix_b

    elif (matrix_a.dtype != np.float64 or matrix_b.dtype != np.float64) and cast:
        debug_print("Recasting matrix data types {a} and {b} to np.float64".format(a=matrix_a.dtype, b=matrix_b.dtype))
        return _cast_to_float64(matrix_a), _cast_to_float64(matrix_b)

    elif matrix_a.dtype != np.float64 or matrix_b.dtype != np.float64:
        err_msg = "Matrix data types must be in concordance; {a} and {b} provided".format(a=matrix_a.dtype,
                                                                                          b=matrix_b.dtype)
        raise ValueError(err_msg)


def _out_matrix(shape, dtype, order="C", out_arr=None, out_t=False):
    """
    Create an all-zero matrix or check to make sure that the provided output array matches

    :param shape: Required output shape
    :type shape: tuple(int)
    :param dtype: Required output data type
    :type dtype: np.dtype
    :param order: Array order (row or column-major)
    :type order: str
    :param out_arr: Provided output array
    :type out_arr: np.ndarray
    :param out_t: Out array has been transposed 
    :type out_t: bool
    :return: Array
    :rtype: np.ndarray
    """

    out_t = False if out_t is None else out_t

    # If there's no output array allocate a new array and return it
    if out_arr is None:
        return np.zeros(shape, dtype=dtype, order=order)

    # Check and make sure the order is correct
    # Note 1d arrays have both flags set
    _order_match = out_arr.flags['C_CONTIGUOUS'] if order == "C" else out_arr.flags['F_CONTIGUOUS']

    # If there are any incompatible parameters, raise an error with the provided and required array parameters
    # Flip them if out_T is set so that the original values and the values which would have to be provided are correct
    if shape != out_arr.shape or dtype != out_arr.dtype or not _order_match or not out_arr.data.contiguous:
        if not out_t or out_arr.ndim == 1:
            _err_shape, _req_shape = out_arr.shape, shape
            _err_order, _req_order = "C" if out_arr.flags['C_CONTIGUOUS'] else "F", order
        else:
            _err_shape, _req_shape = out_arr.shape[::-1], shape[::-1]
            _err_order = "F" if out_arr.flags['C_CONTIGUOUS'] and not out_arr.flags['F_CONTIGUOUS'] else "C"
            _req_order = "C" if order == "F" else "F"

        try:
            _req_dtype = dtype.__name__
        except AttributeError:
            _req_dtype = dtype.name

        _err_msg = "Provided out array is "
        _err_msg += "{s} {d} [{o}_{c}]".format(s=_err_shape, d=out_arr.dtype, o=_err_order,
                                               c="CONTIGUOUS" if out_arr.data.contiguous else "NONCONTIGUOUS")

        _err_msg += "; product requires {s} {d} [{o}_{c}]".format(s=_req_shape, d=_req_dtype, o=_req_order,
                                                                  c="CONTIGUOUS")
        raise ValueError(_err_msg)

    else:
        return out_arr


def _is_dense_vector(m_or_v):
    return not _spsparse.issparse(m_or_v) and ((m_or_v.ndim == 1) or ((m_or_v.ndim == 2) and min(m_or_v.shape) == 1))


def _is_double(arr):
    """
    Return true if the array is doubles, false if singles, and raise an error if it's neither.

    :param arr:
    :type arr: np.ndarray, scipy.sparse.spmatrix
    :return:
    :rtype: bool
    """

    # Figure out which dtype for data
    if arr.dtype == np.float32:
        return False
    elif arr.dtype == np.float64:
        return True
    else:
        raise ValueError("Only float32 or float64 dtypes are supported")


def _is_allowed_sparse_format(matrix):
    """
    Return True if the matrix is dense or a sparse format we can turn into an MKL object. False otherwise.
    :param matrix:
    :return:
    :rtype: bool
    """
    if _spsparse.isspmatrix(matrix):
        return _spsparse.isspmatrix_csr(matrix) or _spsparse.isspmatrix_csc(matrix) or _spsparse.isspmatrix_bsr(matrix)
    else:
        return True


def _empty_output_check(matrix_a, matrix_b):
    """Check for trivial cases where an empty array should be produced"""

    # One dimension is zero
    if min([*matrix_a.shape, *matrix_b.shape]) == 0:
        return True

    # The sparse array is empty
    elif _spsparse.issparse(matrix_a) and min(matrix_a.data.size, matrix_a.indices.size) == 0:
        return True
    elif _spsparse.issparse(matrix_b) and min(matrix_b.data.size, matrix_b.indices.size) == 0:
        return True

    # Neither trivial condition
    else:
        return False


def _validate_dtype():
    """
    Test to make sure that this library works by creating a random sparse array in CSC format,
    then converting it to CSR format and making sure is has not raised an exception.

    """

    test_array = _spsparse.random(5, 5, density=0.5, format="csc", dtype=np.float32, random_state=50)
    test_comparison = test_array.A

    csc_ref, precision_flag = _create_mkl_sparse(test_array)

    try:
        csr_ref = _convert_to_csr(csc_ref)
        final_array = _export_mkl(csr_ref, precision_flag)
        if not np.allclose(test_comparison, final_array.A):
            raise ValueError("Match failed after matrix conversion")
        _destroy_mkl_handle(csr_ref)
    finally:
        _destroy_mkl_handle(csc_ref)


def _empirical_set_dtype():
    """
    Define dtypes empirically
    Basically just try with int64s and if that doesn't work try with int32s
    There's a way to do this with intel's mkl helper package but I don't want to add the dependency
    """
    MKL._set_int_type(_ctypes.c_longlong, np.int64)

    try:
        _validate_dtype()
    except ValueError as err:

        MKL._set_int_type(_ctypes.c_int, np.int32)

        try:
            _validate_dtype()
        except ValueError:
            raise ImportError("Unable to set MKL numeric type")


if MKL.MKL_INT is None:
    _empirical_set_dtype()

if _sklearn_env is not None:
    os.environ['KMP_INIT_AT_FORK'] = _sklearn_env



from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, sparse_matrix_t, _create_mkl_sparse, debug_print, debug_timer,
                                           _export_mkl, _order_mkl_handle, _destroy_mkl_handle, _type_check,
                                           _empty_output_check, _sanity_check, _is_allowed_sparse_format,
                                           _check_return_value)
import ctypes as _ctypes
import numpy as np
import scipy.sparse as _spsparse
from scipy.sparse import isspmatrix_csr as is_csr, isspmatrix_csc as is_csc, isspmatrix_bsr as is_bsr


def _matmul_mkl(sp_ref_a, sp_ref_b):
    """
    Dot product two MKL objects together and return a handle to the result

    :param sp_ref_a: Sparse matrix A handle
    :type sp_ref_a: sparse_matrix_t
    :param sp_ref_b: Sparse matrix B handle
    :param sp_ref_b: sparse_matrix_t
    :return: Sparse matrix handle that is the dot product A * B
    :rtype: sparse_matrix_t
    """

    ref_handle = sparse_matrix_t()

    ret_val = MKL._mkl_sparse_spmm(_ctypes.c_int(10),
                                   sp_ref_a,
                                   sp_ref_b,
                                   _ctypes.byref(ref_handle))

    # Check return
    _check_return_value(ret_val, "mkl_sparse_spmm")
    return ref_handle


def _matmul_mkl_dense(sp_ref_a, sp_ref_b, output_shape, double_precision):
    """
    Dot product two MKL objects together into a dense numpy array and return the result

    :param sp_ref_a: Sparse matrix A handle
    :type sp_ref_a: sparse_matrix_t
    :param sp_ref_b: Sparse matrix B handle
    :type sp_ref_b: sparse_matrix_t
    :param output_shape: The shape of the output array
    This must be correct or the preallocated numpy array won't be correct and this will probably segfault
    :type output_shape: tuple(int, int)
    :param double_precision: The resulting array will be float64
    :type double_precision: bool

    :return: Dense numpy array that's the output of A dot B
    :rtype: np.array
    """

    # Allocate an array for outputs and set functions and types for float or doubles
    output_arr = np.zeros(output_shape, dtype=np.float64 if double_precision else np.float32)
    output_ctype = _ctypes.c_double if double_precision else _ctypes.c_float
    func = MKL._mkl_sparse_d_spmmd if double_precision else MKL._mkl_sparse_s_spmmd

    ret_val = func(10,
                   sp_ref_a,
                   sp_ref_b,
                   101,
                   output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
                   output_shape[1])

    # Check return
    _check_return_value(ret_val, func.__name__)

    return output_arr


def _sparse_dot_sparse(matrix_a, matrix_b, cast=False, reorder_output=False, dense=False):
    """
    Multiply together two scipy sparse matrixes using the intel Math Kernel Library.
    This currently only supports float32 and float64 data

    :param matrix_a: Sparse matrix A in CSC/CSR format
    :type matrix_a: scipy.sparse.spmatrix
    :param matrix_b: Sparse matrix B in CSC/CSR format
    :type matrix_b: scipy.sparse.spmatrix
    :param cast: Should the data be coerced into float64 if it isn't float32 or float64
    If set to True and any other dtype is passed, the matrix data will be modified in-place
    If set to False and any dtype that isn't float32 or float64 is passed, a ValueError will be raised
    Defaults to False
    :param reorder_output: Should the array indices be reordered using MKL
    If set to True, the object in C will be ordered and then exported into python
    If set to False, the array column indices will not be ordered.
    The scipy sparse dot product does not yield ordered column indices so this defaults to False
    :type reorder_output: bool
    :param dense: Should the matrix multiplication yield a dense numpy array
    This does not require any copy and is memory efficient if the output array density is > 50%
    :type dense: bool
    :return: Sparse matrix that is the result of A * B in CSR format
    :rtype: scipy.sparse.csr_matrix
    """

    # Check for allowed sparse matrix types
    if not _is_allowed_sparse_format(matrix_a) or not _is_allowed_sparse_format(matrix_b):
        raise ValueError("Input matrices to dot_product_mkl must be CSR, CSC, or BSR; COO is not supported")

    if is_csr(matrix_a):
        default_output, output_type = _spsparse.csr_matrix, "csr"
    elif is_csc(matrix_a):
        default_output, output_type = _spsparse.csc_matrix, "csc"
    elif is_bsr(matrix_a):
        default_output, output_type = _spsparse.bsr_matrix, "bsr"
    else:
        raise ValueError("Input matrices to dot_product_mkl must be CSR, CSC, or BSR; COO is not supported")

    # Override output if dense flag is set
    default_output = default_output if not dense else np.zeros

    # Check to make sure that this multiplication can work and check dtypes
    _sanity_check(matrix_a, matrix_b)

    # Check for edge condition inputs which result in empty outputs
    if _empty_output_check(matrix_a, matrix_b):
        debug_print("Skipping multiplication because A (dot) B must yield an empty matrix")
        final_dtype = np.float64 if matrix_a.dtype != matrix_b.dtype or matrix_a.dtype != np.float32 else np.float32
        return default_output((matrix_a.shape[0], matrix_b.shape[1]), dtype=final_dtype)

    # Check dtypes
    matrix_a, matrix_b = _type_check(matrix_a, matrix_b, cast=cast)

    t = debug_timer()

    # Create intel MKL objects
    mkl_a, a_dbl = _create_mkl_sparse(matrix_a)
    mkl_b, b_dbl = _create_mkl_sparse(matrix_b)

    t = debug_timer("Created MKL sparse handles", t)

    # Call spmmd for dense output directly if the dense flag is set
    if dense:
        dense_arr = _matmul_mkl_dense(mkl_a, mkl_b, (matrix_a.shape[0], matrix_b.shape[1]), a_dbl or b_dbl)

        debug_timer("Multiplied matrices", t)

        _destroy_mkl_handle(mkl_a)
        _destroy_mkl_handle(mkl_b)

        return dense_arr

    # Call spmm for sparse output if the dense flag is not set and then export the sparse matrix to python
    else:
        # Dot product
        mkl_c = _matmul_mkl(mkl_a, mkl_b)

        _destroy_mkl_handle(mkl_a)
        _destroy_mkl_handle(mkl_b)

        t = debug_timer("Multiplied matrices", t)

        # Reorder
        if reorder_output:
            _order_mkl_handle(mkl_c)

            t = debug_timer("Reordered output indices", t)

        # Extract
        python_c = _export_mkl(mkl_c, a_dbl or b_dbl, output_type=output_type)
        _destroy_mkl_handle(mkl_c)

        debug_timer("Created python handle", t)

        return python_c



#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from tvtk.api import tvtk

from mayavi.sources.vtk_data_source import VTKDataSource

from mayavi.modules.glyph import Glyph

import numpy as np

# -----------------------------------------------------------------------------

from wsurf.geolab.plotter import plotutilities

# -----------------------------------------------------------------------------

'''pointsource.py: Point plot source class, for meshes and arrays'''

__author__ = 'Davide Pellis'


#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                                   Points
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------


class Points(object):
    '''The plot source for points.

    If added to a list, it can be displayed with geolab.plotter.view()

    Parameters
    ----------
    points : np.array, (V, 3)
        An array with the x,y,z coordinates of the points to plot.

    **kwargs
        point_data : np.array, (V,)
            An array containing additional data corresponding to the points.
            The i-th entry of the array is associate with the i-th vertex row.
        vertex_indices : np.array of int, (N,)
            If specified, only the vertices with index contained in this
            array will be plotted.
        color : str / tuple / list of str and tuple
            The color of the plotted object. Colors can be specified as:
                - `(r,g,b)` tuples, with `r`, `g`, `b` in the interval [0,1]
                - a keyword color
                - a keyword for a color Lookup Table (LUT)
                - a list of `(r,g,b)` tuples and keyword colors
            Default is `cornflower`. If a keyword for a color Lookup Table
            (LUT) is specified, the points are colored according to a
            `point_data` -> color mapping.
        opacity : float / list of float
            a [0,1] float / a list of [0,1] floats of the same length of colors
        lut_range : str / list
            A keyword that specifies the range of the color Lookup Table
            (LUT) for data display.
                - `'-:0:+'` for symmetric LUT range, zero is the central color
                - `'0:+'` for positive LUT range, zero is the first color
                - `'-:0'` for negative LUT range, zero is the last color
                - `'-:+'` for min-max LUT range
                - `[min,max]` for user defined LUT range
            Default is `'-:0:+'`.
        glyph_type: str
            Specifies how the points are displayed. It can be
            `sphere`, `cube`, `axes`, `cone`, `wireframe`.
            Default is `sphere`

    '''

    def __init__(self, points, **kwargs):
        plotutilities.check_arguments(**kwargs)

        self._points          = points

        self._geometry        = None

        self._data_range      = None

        self._vertex_data     = kwargs.get('vertex_data',    None)

        self._indices         = kwargs.get('vertex_indices', None)

        self._color           = kwargs.get('color',          'cornflower')

        self._opacity         = kwargs.get('opacity',        1)

        self._line_width      = kwargs.get('line_width',     1)

        self._scale_factor    = kwargs.get('scale_factor',   1)

        self._glyph_type      = kwargs.get('glyph_type',     'sphere')

        self._radius          = kwargs.get('radius',         0.1)

        self._height          = kwargs.get('height',         0.1)

        self._resolution      = kwargs.get('resolution',     20)

        self._lut_range       = kwargs.get('lut_range',      '-:0:+')

        self._lut_expansion   = kwargs.get('lut_expansion',  2)

        self._reverse_lut     = kwargs.get('reverse_lut',    False)

        self._shading         = kwargs.get('shading',        True)

        self._glossy          = kwargs.get('glossy',         0.3)

        self._scaling         = kwargs.get('scaling',        False)

        self._data_points     = kwargs.get('data_points',    None)

        self.name             = kwargs.get('name',           'points')

    #--------------------------------------------------------------------------
    #                                 Pipeline
    #--------------------------------------------------------------------------
    # - .source =   VTKDataSource
    # - .data   =     \_Unstructured_grid
    # - .module =        \_Module_manager
    # - .glyph  =           \__Glyph
    #--------------------------------------------------------------------------

        self._make_data()

        self._module = plotutilities.make_module(self._color,
                                                 self._opacity,
                                                 self._lut_range,
                                                 self._lut_expansion,
                                                 self._reverse_lut,
                                                 self._data_range)

        self._make_glyph()

        self._assemble_pipeline()

        self.on_scene = False

    @property
    def type(self):
        return 'Points-source'

    #--------------------------------------------------------------------------
    #                               Data Structure
    #--------------------------------------------------------------------------

    def _make_data(self):
        try:
            self._points.shape
        except AttributeError:
            self._points = np.array(self._points)
        if len(self._points.shape) == 1:
            self._points = np.array([self._points])
        if self._indices is not None:
            points = self._points[self._indices]
        else:
            points = self._points
        self._data = tvtk.PolyData(points=points)
        if self._data_points is not None:
            V = len(self._data_points)
            self._data.point_data.scalars = np.arange(V)
            self._data_range = [0, V]
        elif self._vertex_data is None:
            V = self._points.shape[0]
            self._data.point_data.scalars = np.zeros(V)
            self._data_range = [0, 0]
        else:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data_range = [np.min(scalars), np.max(scalars)]
        self._data.point_data.scalars.name = self.name

    # -------------------------------------------------------------------------
    #                                Glyph
    # -------------------------------------------------------------------------

    def _make_glyph(self):
        self._glyph = Glyph()
        if self._scaling:
            self._glyph.glyph.scale_mode = 'scale_by_scalar'
        else:
            self._glyph.glyph.scale_mode = 'data_scaling_off'
        self._glyph.actor.property.opacity = self._opacity
        #print(glyph.glyph.glyph_source.glyph_dict)

        if self._glyph_type == 'wireframe':
            self._glyph.actor.actor.property.representation = 'points'
            self._glyph.actor.actor.property.point_size = 1
            self._glyph.actor.actor.property.render_points_as_spheres = True
            g_src = self._glyph.glyph.glyph_source.glyph_dict['glyph_source2d']
            self._glyph.glyph.glyph_source.glyph_source.dash = True
            self._glyph.glyph.glyph_source.glyph_source.glyph_type = 'none'
            self._glyph.glyph.glyph_source.glyph_source.scale = 0
            self._glyph.actor.property.line_width = self._line_width

        elif self._glyph_type == 'cube':
            self._glyph.actor.actor.property.representation = 'surface'
            g_src = self._glyph.glyph.glyph_source.glyph_dict['cube_source']
            g_src.x_length = 2*self._radius
            g_src.y_length = 2*self._radius
            g_src.z_length = 2*self._radius

        elif self._glyph_type == 'sphere':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['sphere_source']
            g_src.radius = self._radius
            g_src.phi_resolution = self._resolution
            g_src.theta_resolution = self._resolution

        elif self._glyph_type == 'cone':
            self._glyph.actor.actor.property.representation = 'surface'
            g_src = self._glyph.glyph.glyph_source.glyph_dict['cone_source']
            g_src.resolution = self._resolution
            g_src.radius = self._radius
            g_src.height = self._height
            g_src.direction =  np.array([0, 0, 1])
            g_src.center = np.array([0, 0, -self._height/2])


        elif self._glyph_type == 'axes':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['axes']
            self._glyph.actor.property.line_width = self._line_width

        self._glyph.glyph.glyph_source.glyph_source = g_src
        self._glyph.glyph.glyph.scale_factor = self._scale_factor

        if not self._shading:
            self._glyph.actor.actor.property.lighting = False

        self._glyph.actor.actor.property.specular = min(self._glossy, 1)
        sp = 21. * self._glossy + 0.001
        self._glyph.actor.actor.property.specular_power = sp

    #--------------------------------------------------------------------------
    #                                Pipeline
    #--------------------------------------------------------------------------

    def _assemble_pipeline(self):
        src = VTKDataSource(data=self._data)
        self._module.add_child(self._glyph)
        src.add_child(self._module)
        self.source = src

    # -------------------------------------------------------------------------
    #                                 Update
    # -------------------------------------------------------------------------

    def _update_data(self, **kwargs):
        self._points = kwargs.get('points', self._points)
        self._indices = kwargs.get('vertex_indices', self._indices)
        try:
            self._points.shape
        except AttributeError:
            self._points = np.array(self._points)
        if len(self._points.shape) == 1:
            self._points = np.array([self._points])
        if self._indices is not None:
            points = self._points[self._indices]
        else:
            points = self._points
        self._data.set(points=points)
        self._vertex_data = kwargs.get('vertex_data', self._vertex_data)
        if self._vertex_data is not None:
            if self._points.shape[0] != self._vertex_data.shape[0]:
                self._vertex_data = None
        if self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        else:
            V = points.shape[0]
            self._data.point_data.scalars = np.zeros(V)
            self._data.point_data.scalars.name = self.name
            self._data_range = [0,0]
        self._data.modified()

    def _update_lut_range(self, **kwargs):
        self._lut_range = kwargs.get('lut_range', self._lut_range)
        lut_range = plotutilities.make_lut_range(self._lut_range, self._data_range)
        self._module.scalar_lut_manager.data_range = lut_range

    def _update_glossy(self, **kwargs):
        self._glossy = kwargs.get('glossy', self._glossy)
        self._glyph.actor.actor.property.specular = min(self._glossy, 1)
        sp = 21. * self._glossy + 0.001
        self._glyph.actor.actor.property.specular_power = sp

    def _update_radius(self, **kwargs):
        g_src = self._glyph.glyph.glyph_source.glyph_source
        if self._glyph_type == 'cube':
            g_src.x_length = 2*self._radius
            g_src.y_length = 2*self._radius
            g_src.z_length = 2*self._radius
        elif self._glyph_type == 'sphere':
            self._radius = kwargs.get('radius', self._radius)
            g_src.radius = self._radius

    def _update_color(self, **kwargs):
        if 'color' in kwargs:
            if kwargs['color'] is not self._color:
                self._color = kwargs['color']
                plotutilities.make_lut_table(self._module,
                                             self._color,
                                             self._opacity,
                                             self._lut_expansion,
                                             self._reverse_lut)


    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        self._update_data(**kwargs)
        self._update_lut_range(**kwargs)
        self._update_color(**kwargs)
        self._update_glossy(**kwargs)
        self._update_radius(**kwargs)
        # --------------------------- bug fix ---------------------------------
        # plot 2 times for updation of cone center
        if self._glyph_type == 'cone':
            g_src = self._glyph.glyph.glyph_source.glyph_source
            g_src.center = np.array([0, 0, -self._height/2])
        # ---------------------------------------------------------------------
        self.source.update()


    def kwargs(self):
        kwargs = {}
        kwargs['name'] = self.name
        if hasattr(self._points, 'vertices'):
            if self._indices is None:
                kwargs['points'] = np.copy(self._points.vertices)
            else:
                kwargs['points'] = np.copy(self._points.vertices[self._indices])
        else:
            kwargs['points'] = np.copy(self._points)
        kwargs['vertex_data'] = np.copy(self._vertex_data)
        return kwargs

    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------

'''
def _update_data(self, **kwargs):

        if 'points' in kwargs:
            points = kwargs['points']
            if type(points) == list:
                points = np.array(points)
            if hasattr(points, 'shape'):
                if len(points.shape) == 1:
                    points = np.array([points])
            elif points.type == 'Mesh':
                self._indices = kwargs.get('indices', self._indices)
                if self._indices is not None:
                    if type(self._indices) == list:
                        self._indices = np.array(self._indices)
                    points = self._geometry.vertices[self._indices,:]
                else:
                    points = self._geometry.vertices
            self._points = points
        else:
            points = self._points
        print(points)
        self._data.set(points=points)
        self._vertex_data = kwargs.get('vertex_data', self._vertex_data)
        if self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        else:
            V = points.shape[0]
            self._data.point_data.scalars = np.zeros(V)
            self._data.point_data.scalars.name = self.name
            self._data_range = [0,0]
        self._data.modified()
'''





try:
    from wsurf.geolab.optimization.sparse_mkl.scipy_aliases import spsolve
except:
    from scipy.sparse.linalg import spsolve
    print('Efficiency Warning: mkl spsolve not available')

try:
    from wsurf.geolab.optimization.sparse_mkl.sparse_dot import dot_product_mkl as spdot
    from wsurf.geolab.optimization.routines import hdot_mkl as hdot

except:
    from wsurf.geolab.optimization.routines import spdot
    from wsurf.geolab.optimization.routines import hdot
    print('Efficiency Warning: mkl dot not available')

from wsurf.geolab.optimization.guidedprojectionbase import GuidedProjectionBase





from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, _type_check, _sanity_check, _empty_output_check, _get_numpy_layout,
                                           LAYOUT_CODE_C, LAYOUT_CODE_F, _out_matrix, debug_print)

import numpy as np
import ctypes as _ctypes


def _dense_matmul(matrix_a, matrix_b, double_precision, scalar=1., out=None, out_scalar=None):

    # Reshape matrix_b to a column instead of a vector if it's 1d
    flatten_output = matrix_b.ndim == 1
    matrix_b = matrix_b.reshape(-1, 1) if flatten_output else matrix_b

    # Get dimensions
    m, n, k = matrix_a.shape[0], matrix_b.shape[1], matrix_a.shape[1]
    output_shape = (m, n)

    # Set the MKL function for precision
    func = MKL._cblas_dgemm if double_precision else MKL._cblas_sgemm

    # Get the memory order for arrays
    layout_a, ld_a = _get_numpy_layout(matrix_a)
    layout_b, ld_b = _get_numpy_layout(matrix_b)

    # If they aren't the same, use the order for matrix a and have matrix b transposed
    op_b = 112 if layout_b != layout_a else 111

    # Set output array; use the memory order from matrix_a
    out_order, ld_out = ("C", output_shape[1]) if layout_a == LAYOUT_CODE_C else ("F", output_shape[0])

    # Allocate an array for outputs and set functions and types for float or doubles
    output_arr = _out_matrix(output_shape, np.float64 if double_precision else np.float32, order=out_order, out_arr=out)
    output_ctype = _ctypes.c_double if double_precision else _ctypes.c_float

    func(layout_a,
         111,
         op_b,
         m,
         n,
         k,
         scalar,
         matrix_a,
         ld_a,
         matrix_b,
         ld_b,
         float(out_scalar) if out_scalar is not None else 1.,
         output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
         ld_out)

    return output_arr.ravel() if flatten_output else output_arr


def _dense_dot_dense(matrix_a, matrix_b, cast=False, scalar=1., out=None, out_scalar=None):

    _sanity_check(matrix_a, matrix_b, allow_vector=True)

    # Check for edge condition inputs which result in empty outputs
    if _empty_output_check(matrix_a, matrix_b):
        debug_print("Skipping multiplication because A (dot) B must yield an empty matrix")
        final_dtype = np.float64 if matrix_a.dtype != matrix_b.dtype or matrix_a.dtype != np.float32 else np.float32
        return _out_matrix((matrix_a.shape[0], matrix_b.shape[1]), final_dtype, out_arr=out)

    matrix_a, matrix_b = _type_check(matrix_a, matrix_b, cast=cast)

    a_dbl, b_dbl = matrix_a.dtype == np.float64, matrix_b.dtype == np.float64

    return _dense_matmul(matrix_a, matrix_b, a_dbl or b_dbl, scalar=scalar, out=out, out_scalar=out_scalar)



import numpy as np

from scipy.sparse import find

from geolab.plotter.glyphs import MeshGlyph

from geolab.geometry import mesh_plane

from geolab.plotter import Faces

from geolab.plotter import Edges

from geolab.plotter import view

# -----------------------------------------------------------------------------


def plot_matrix(matrix):
    F1 = matrix.shape[0]
    F2 = matrix.shape[1]
    x = np.linspace(0, 1000, F1+1)
    y = np.linspace(0, 1000*F2/F1, F2+1)
    i, j, d = find(matrix)
    N = len(i)
    F = np.reshape(np.arange(4*N), (N, 4), 'C')
    Vx = np.column_stack((x[i], x[i+1], x[i+1], x[i]))
    Vy = np.column_stack((y[j], y[j], y[j+1], y[j+1]))
    V = np.column_stack((Vy.flatten('C'), -Vx.flatten('C'), np.zeros(4*N)))
    M = MeshGlyph(V, quads=F)
    source = [Faces(M, face_data=np.abs(d), color='plasma', lut_range='0:+')]
    B = mesh_plane(1, 1, (0, 1000*F2/F1), (-1000, 0))
    source.append(Edges(B, color=(200, 200, 200)))
    view(source, background=(1, 1, 1))
    return source




from setuptools import find_packages, setup

setup(
    name='wsurf',
    version='0.0.01',
    description='',
    long_description='',
    url='https://github.com/DavidePellis/wsurf',
    author='Davide Pellis',
    download_url='https://github.com/DavidePellis/wsurf/archive/0.0.01.tar.gz',
    author_email='davidepellis@gmail.com',
    packages=find_packages(),
    package_data={'': ['icons/*.png', '*.obj']},
    classifiers=['Development Status :: 1 - Planning'],
    license='MIT',
)


#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from pyface.image_resource import ImageResource

from traits.api import HasTraits, Instance, Property, Enum, Button,String,\
                       on_trait_change, Float, Bool, Int,Constant, ReadOnly,\
                       List, Array, Range, Str, Color

from traitsui.api import View, Item, HSplit, VSplit, InstanceEditor, HGroup,\
                         Group, ListEditor, Tabbed, VGroup, CheckListEditor,\
                         ArrayEditor, Action, ToolBar, Separator,EnumEditor,\
                         ListStrEditor, ColorEditor, Controller

#------------------------------------------------------------------------------

from wsurf.geolab.plotter.bsplinesurfaceplotmanager import BsplineSurfacePlotManager

#------------------------------------------------------------------------------

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                               Geolab Mesh
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class GMHandler(Controller):

    def close(self, info, is_ok):
        info.object._closed = True
        Controller.close(self, info, is_ok)
        return True


class GeolabBSplineSurface(BsplineSurfacePlotManager):

    _handler = GMHandler()

    _closed = Bool(True)

    plot_scale = Range(-10, 10, 0)

    surface_glossy_range = Range(0., 1., 0.5)

    control_color_select = Color((10, 10, 10), width=20, label='control_color')

    surface_color_select = Color((122, 163, 230), width=20, label='surface_color')

    control_plot = Str('wireframe', label='Control polygon')

    control_callbacks = ['color', 'wireframe', 'none']

    surface_plot = Str('color', label='Surface')

    surface_callbacks = ['color', 'gaussian curv.', 'mean curv.', 'none']

    surface_data_range = Range(0.00, 1.00, 0.80)

    show_edge_legend = Bool(False, label='legend')

    show_face_legend = Bool(False, label='legend')

    view = View(
                VGroup(VGroup(
                       Item('plot_scale',
                            resizable=True),
                       show_border=True),
                       VGroup(
                       HGroup(VGroup(Item('control_plot',
                                          resizable=True,
                                          editor=CheckListEditor(values=
                                                          control_callbacks),),
                                     Item('surface_plot',
                                          resizable=True,
                                          editor=CheckListEditor(values=
                                                          surface_callbacks),),
                                     ),
                              VGroup(Item('control_color_select',
                                          show_label=False,
                                          resizable=False,
                                          enabled_when='control_plot == "color"',
                                          editor=ColorEditor(),),
                                     Item('surface_color_select',
                                          show_label=False,
                                          resizable=False,
                                          enabled_when='surface_plot == "color"',
                                          editor=ColorEditor(),),
                                     ),
                              VGroup('show_edge_legend',
                                     '_',
                                     'show_face_legend',
                                     ),
                               ),
                       show_border=True),
                       VGroup(
                       HGroup(Item('surface_glossy_range',
                                   resizable=True),
                              ),
                       HGroup(Item('surface_data_range',
                                   resizable=True),
                               ),
                       show_border=True),
                       show_border=True),
                handler=_handler,
                title='B-spline surface plot settings',
                icon=ImageResource('../icons/plotmanager.png'),)

    def __init__(self, *args):
        BsplineSurfacePlotManager.__init__(self, *args)

        self._control_callbacks = {'color': self.plot_plain_edges,
                                'wireframe': self.plot_wireframe_control_polygon,
                                'none': self.hide_control_polygon}

        self._surface_callbacks = {'color': self.plot_plain_surface,
                                'gaussian curv.': self.plot_gaussian_curvature,
                                'mean curv.': self.plot_mean_curvature,
                                'none': self.hide_surface}

    def start(self):
        if not self._closed:
            pass
            self._handler.info.ui.dispose()
        self.configure_traits()
        self._closed = False

    def close(self):
        try:
            self._closed = True
            self._handler.info.ui.dispose()
        except:
            pass

    #--------------------------------------------------------------------------
    #                                    Plot
    #--------------------------------------------------------------------------

    def add_surface_callback(self, callback, name):
        self._surface_callbacks[name] = callback
        if name not in self.surface_callbacks:
            self.surface_callbacks.append(name)

    def add_edge_callback(self, callback, name):
        self._edge_callbacks[name] = callback
        if name not in self.edge_callbacks:
            self.edge_callbacks.append(name)

    @on_trait_change('control_plot')
    def edge_plot_change(self):
        self.remove_control_polygon()
        self.remove_edge_sources()
        self.update_plot()

    @on_trait_change('surface_plot, smooth_surface')
    def face_plot_changed(self):
        self.remove_surface()
        self.remove_face_sources()
        self.update_plot()

    def core_plot(self, **kwargs):
        control_plot = self._control_callbacks[self.control_plot]
        control_plot()
        surface_plot = self._surface_callbacks[self.surface_plot]
        surface_plot()
        self.make_face_legend()
        BsplineSurfacePlotManager.core_plot(self)


    #--------------------------------------------------------------------------
    #                             Predefined Plots
    #--------------------------------------------------------------------------

    def plot_plain_edges(self):
        self.plot_control_polygon()

    def plot_wireframe_control_polygon(self):
        self.plot_control_polygon(color='k', tube_radius=None)

    def plot_plain_surface(self):
        self.plot_surface(glossy=self.surface_glossy_range)

    def plot_gaussian_curvature(self):
        K, H, k1, k2 = self.bspline.curvatures()
        vertex_data = K
        val = np.max(np.abs(vertex_data)) * self.surface_data_range
        self.plot_surface(vertex_data = vertex_data,
                          glossy=self.surface_glossy_range,
                          color='turbo',
                          lut_range=[-val, val])

    def plot_mean_curvature(self):
        K, H, k1, k2 = self.bspline.curvatures()
        vertex_data = H
        val = np.max(np.abs(vertex_data)) * self.surface_data_range
        self.plot_surface(vertex_data=vertex_data,
                          glossy=self.surface_glossy_range,
                          color='turbo',
                          lut_range=[-val, val])

    @on_trait_change('show_face_legend')
    def make_face_legend(self):
        try:
            if self.show_face_legend:
                self.get_source('faces').show_legend(label=self.face_plot)
            else:
                self.get_source('faces').hide_legend()
        except:
            pass

    # -------------------------------------------------------------------------
    #                                 Select
    # -------------------------------------------------------------------------

    def select_off(self):
        super(GeolabBSplineSurface, self).select_off()

    # -------------------------------------------------------------------------
    #                                 Scales
    # -------------------------------------------------------------------------

    @on_trait_change('surface_data_range, surface_glossy_range')
    def plot_updated(self):
        self.update_plot()

    @on_trait_change('plot_scale')
    def set_scale(self):
        #self.handler.set_state(None)
        self.scale = 4**(0.1*self.plot_scale)
        self.glyph_scale = 4**(0.1*self.plot_scale)
        self.update_plot()

    @on_trait_change('surface_color_select, control_color_select')
    def set_color(self):
        ec = self.control_color_select.getRgb()
        fc = self.surface_color_select.getRgb()
        self.control_color = (ec[0], ec[1], ec[2])
        self.surface_color = (fc[0], fc[1], fc[2])
        self.update_plot()





from scipy import sparse


def spdot(A, B):
    return sparse.spmatrix.dot(A, B)


def hdot(A):
    return sparse.spmatrix.dot(A, A.transpose())


def hdot_mkl(A):
    from geolab.optimization.sparse_mkl.sparse_dot import gram_matrix_mkl
    B = gram_matrix_mkl(A)
    B = B + sparse.triu(B, 1).transpose()
    return B



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 17 20:15:01 2018

@author: Davide
"""

from wsurf.geolab.gui.handler import Handler

from wsurf.geolab.gui.geolabcomponent import GeolabComponent

from wsurf.geolab.gui.geolabgui import GeolabGUI

from wsurf.geolab.gui.geolabgui import Viewer

from wsurf.geolab.gui.scenemanager import SceneManager

from wsurf.geolab.gui.multiscenemanager import MultiSceneManager